/***************************************************************************
  $RCSfile: account.h,v $
  -------------------
  cvs         : $Id: account.h,v 1.10 2004/01/25 18:51:22 aquamaniac Exp $
  begin       : Tue Oct 14 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifndef AQMONEY_ACCOUNT_H
#define AQMONEY_ACCOUNT_H

#include <openhbci2.h>
#include <openhbci2/value.h>
#include <openhbci2/date.h>

using namespace HBCI;
#include <string>
using namespace std;

class Account;

#define AQM_TRANSACTION_MATCH_POSITIVE   0x0001
#define AQM_TRANSACTION_MATCH_NEGATIVE   0x0002

#define AQM_TRANSACTION_MATCH_DEFAULT \
    AQM_TRANSACTION_MATCH_POSITIVE | \
    AQM_TRANSACTION_MATCH_NEGATIVE


#include <aqmoney2/aqmapi.h>
#include <aqmoney2/storage.h>
#include <gwenhywfar/db.h>



/**
 *
 */
class Account {
public:
  /**
   * @short Stores all data belonging to one specific balance.
   *
   * This class stores the data that represent one specific balance. An
   * account can have multiple balances according to different balancing
   * rules.  */
  class Balance {
  private:
    Value _value;
    Date _date;
    Time _time;
  public:
    /** @name Constructors and Destructor
     *
     * The actual balance. */
    /*@{*/
    Balance();
    Balance(const Value &value,
            const Date &date,
            const Time &tim);
    ~Balance();
    /*@}*/

    /** @name Balance
     *
     * The actual balance. */
    /*@{*/
    /** @short Returns the value of this balance. */
    const Value &value() const { return _value;};
    /**  Set the value of this balance.
     *
     * @param v The value that this balance should be set to. */
    void setValue(const Value &v) { _value=v;};
    /*@}*/

    /** @name Date and Time
     *
     * The point in thime when this balance was created. */
    /*@{*/
    /**  Returns the Date of this balance. 
     *
     * Returns the Date when this balance was valid/computed/fixed. */
    const Date &date() const { return _date;};
    /**  Set the Date of this balance. 
     *
     * @param d The Date at which this balance is valid. */
    void setDate(const Date &d) { _date=d;};

    /**  Returns the Time of day of this balance. */
    const Time &time() const { return _time;};
    /**  Set the Time of this balance.
     * 
     * @param t The Time of this balance. */
    void setTime(const Time &t) { _time=t;};
    /*@}*/

  };

private:
  AqMoneyAPI *_api;
  string _accountId;
  string _subId;
  int _country;
  string _bankCode;
  string _serverBankCode;
  string _accountName;
  string _customer;
  string _currency;
  string _userName;
  Pointer<TransactionStorage> _storage;
  Pointer<TransactionStorage> _transferStorage;
  Balance _bookedBalance;
  Balance _notedBalance;
  Value _bankLine;
  GWEN_DB_NODE *_standingOrders;
  GWEN_DB_NODE *_datedTransfers;
  GWEN_DB_NODE *_notedTransfers;

  GWEN_DB_NODE *_getMatchingTransfers(GWEN_DB_NODE *transfers,
                                      const char *tname,
                                      int country,
                                      const string &bankCode,
                                      const string &accountId);

protected:
  virtual Error fromConfig(GWEN_DB_NODE *db);
public:
  Account(AqMoneyAPI *api,
          const string &accountId,
          const string &subid,
          int country,
          const string &bankCode,
          const string &serverBankCode,
          const string &accountName,
          const string &customer,
          const string &currency,
          const string &userName,
          const string &storageType,
          const string &storageAddr);
  Account(AqMoneyAPI *api, GWEN_DB_NODE *db) throw (Error);
  virtual ~Account();

  Pointer<TransactionStorage> storage() { return _storage;};
  Pointer<TransactionStorage> transferStorage() { return _transferStorage;};

  const string &accountId() {return _accountId;};
  const string &accountSubId() {return _subId;};

  const string &accountName() const { return _accountName;};
  void setAccountName(const string &s) { _accountName=s;};

  int country() const {return _country;};
  const string &bankCode() const { return _bankCode;};
  /**
   * In some cases the bankCode of the account differs from the bank code
   * of the server. When performing jobs on such accounts we need to know
   * both bank codes: The serverBankCode for the communication with the
   * server and the bank code of the account.
   */
  const string &serverBankCode() const { return _serverBankCode;};
  const string &customer() const { return _customer;};
  void setCustomer(const string &s) { _customer=s;};
  const string &currency() const { return _currency;};
  void setCurrency(const string &s) { _currency=s;};
  const string &userName() const { return _userName;};
  void setUserName(const string &u) { _userName=u;};

  const Balance &bookedBalance() const { return _bookedBalance;};
  void setBookedBalance(const Balance &bal) { _bookedBalance=bal;};
  const Balance &notedBalance() const { return _notedBalance;};
  void setNotedBalance(const Balance &bal) { _notedBalance=bal;};
  const Value &creditLine() const { return _bankLine;};
  void setCreditLine(const Value &bl) { _bankLine=bl;};

  Error addStandingOrder(GWEN_DB_NODE *xa);
  void clearStandingOrders();
  GWEN_DB_NODE *getStandingOrders(int country=280,
                                  const string &bankCode="*",
                                  const string &accountId="*");

  Error addDatedTransfer(GWEN_DB_NODE *xa);
  void clearDatedTransfers();
  GWEN_DB_NODE *getDatedTransfers(int country=280,
                                  const string &bankCode="*",
                                  const string &accountId="*");

  Error addNotedTransfer(GWEN_DB_NODE *xa);
  void clearNotedTransfers();
  GWEN_DB_NODE *getNotedTransfers(int country=280,
                                  const string &bankCode="*",
                                  const string &accountId="*");

  virtual Error toConfig(GWEN_DB_NODE *db);

};


#endif

