/***************************************************************************
  $RCSfile: accountjobs.cpp,v $
  -------------------
  cvs         : $Id: accountjobs.cpp,v 1.23 2004/05/27 10:10:44 aquamaniac Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

/* Internationalization */
#ifdef ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define I18N(m) dgettext("aqmoney2",m)
#else
# define I18N(m) m
#endif

#include "accountjobs.h"
#include "aqmstring.h"
#include <gwenhywfar/debug.h>
#include <openhbci2/datetime.h>

AccountJob::AccountJob(const string &name,
                       Pointer<Customer> c,
                       Pointer<Account> acc)
:OutboxJob(name, c, acc.ref().accountId()), _account(acc) {
  assert(c.isValid());
  assert(acc.isValid());
  setProperty("accountsubid", acc.ref().accountSubId());
  setProperty("bankcode", acc.ref().bankCode());
  setIntProperty("country", acc.ref().country());

}



AccountJob::~AccountJob(){
}


const string &AccountJob::serverBankCode() {
  return customer().ref().user().ref().bank().ref().bankCode();
}


Error AccountJob::commit(){
  return Error();
}


Error AccountJob::evaluate(Pointer<Outbox> outbox){
  if (results().empty()) {
    setStatus(HBCI_JOB_STATUS_DONE);
    setResult(HBCI_JOB_RESULT_FAILED);
    return Error("AdminJob::evaluate",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no results");
  }
  return OutboxJob::evaluate(outbox);
}



void AccountJob::storeResult(GWEN_DB_NODE *rn, int code, const string &text){
  GWEN_DB_NODE *n;

  n=GWEN_DB_GetGroup(rn,
                     GWEN_DB_FLAGS_DEFAULT |
                     GWEN_PATH_FLAGS_CREATE_GROUP,
                     "jobresult");
  GWEN_DB_SetIntValue(n, GWEN_DB_FLAGS_DEFAULT,
                      "resultCode",
                      code);
  GWEN_DB_SetCharValue(n, GWEN_DB_FLAGS_DEFAULT,
                       "resultText",
                       text.c_str());
}


string AccountJob::resultString(bool sh){
  list<MsgJob::Result>::const_iterator rit;
  Error err;
  string res;
  Pointer<Account> acc;
  Pointer<Bank> bank;
  bool haveWarnings;

  haveWarnings=false;
  if (!results().empty()) {
    for (rit=results().begin();
         rit!=results().end();
         rit++) {
      if ((*rit).resultCode()>=3000 &&
          (*rit).resultCode()<4000) {
        haveWarnings=true;
        break;
      }
    } // for
  }

  acc=account();
  bank=customer().ref().user().ref().bank();
  res+=AqmoneyString::format(I18N(" %s with account %s at %s"),
                             name().c_str(),
                             acc.ref().accountId().c_str(),
                             bank.ref().name().empty()
                               ? bank.ref().bankCode().c_str()
                               : bank.ref().name().c_str()
                            );
  if (sh) {
    res+=": ";
  }
  else {
    res+="\n";

    res+=I18N("  Overall result: ");
  }

  switch(result()) {
  case HBCI_JOB_RESULT_SUCCESS:        res+=I18N("OK"); break;
  case HBCI_JOB_RESULT_PARTLY_SUCCESS: res+=I18N("partly OK"); break;
  case HBCI_JOB_RESULT_FAILED:         res+=I18N("failed"); break;
  default:                             res+=I18N("unknown"); break;
  } // switch

  if (haveWarnings)
    res+=I18N(" (with warnings)");
  res+="\n";

  if (!sh || haveWarnings) {
    if (!results().empty()) {
      res+=I18N("  Individual results:\n");
      for (rit=results().begin();
           rit!=results().end();
           rit++) {
        res+="   ";
        if ((*rit).resultCode()>=9000)
          res+=I18N("Error  : ");
        else if ((*rit).resultCode()>=3000 && (*rit).resultCode()<4000)
          res+=I18N("Warning: ");
        else
          res+=I18N("Notice : ");
        res+=String::num2string((*rit).resultCode(), false, 4);
        res+=" ";
        res+=(*rit).resultMessage();
        res+="\n";
      } // for result
    }
    else {
      res+=I18N("  No individual results for this job\n");
    }
  }
  return res;
}



AqMoneyAPI *AccountJob::api(){
  try {
    return (AqMoneyAPI*)(customer().ref().user().ref().bank().ref().hbci());
  }
  catch (Error xerr) {
    DBG_INFO(0, "Exception: %s", xerr.errorString().c_str());
    throw Error("AccountJob::api()", xerr);
  }
}



JobGetBalance::JobGetBalance(Pointer<Customer> c,
                             Pointer<Account> acc)
:AccountJob("JOBGetBalance", c, acc) {
  setProperty("allaccounts", "N");
}



JobGetBalance::~JobGetBalance(){
}



Error JobGetBalance::commit(){
  string currency;
  Value val;
  Date d;
  Time t;
  GWEN_DB_NODE *rsp;
  GWEN_DB_NODE *n;

  rsp=responseData();
  if (!rsp)
    return Error("JobGetBalance::evaluate",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no results");
  rsp=GWEN_DB_GetGroup(rsp, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "balance");
  if (!rsp)
    return Error("JobGetBalance::evaluate",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no result for this job, maybe an error occurred ?");

  try {
    currency=GWEN_DB_GetCharValue(rsp, "currency", 0, "EUR");

    // get booked balance
    n=GWEN_DB_GetGroup(rsp, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "booked");
    if (n){
      DBG_INFO(0, "Have booked balance");
      val=Value(GWEN_DB_GetCharValue(n, "value", 0, "0"));
      if (strcasecmp(GWEN_DB_GetCharValue(n, "debitmark", 0, "C"),"D")==0)
        val=Value(-val.getValue(), currency);
      else
        val=Value(val.getValue(), currency);
      d=Date(GWEN_DB_GetCharValue(n, "date", 0, ""));
      t=Time(GWEN_DB_GetCharValue(n, "time", 0, ""));
      account().ref().setBookedBalance(Account::Balance(val, d, t));
    }

    // get noted balance
    n=GWEN_DB_GetGroup(rsp, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "noted");
    if (n){
      DBG_INFO(0, "Have noted balance");
      val=Value(GWEN_DB_GetCharValue(n, "value", 0, "0"));
      if (strcasecmp(GWEN_DB_GetCharValue(n, "debitmark", 0, "C"),"D")==0)
        val=Value(-val.getValue(), currency);
      else
        val=Value(val.getValue(), currency);
      d=Date(GWEN_DB_GetCharValue(n, "date", 0, ""));
      t=Time(GWEN_DB_GetCharValue(n, "time", 0, ""));
      account().ref().setNotedBalance(Account::Balance(val, d, t));
    }

    n=GWEN_DB_GetGroup(rsp, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "creditline");
    if (n){
      DBG_INFO(0, "Have credit line");
      val=Value(GWEN_DB_GetCharValue(n, "value", 0, "0"));
      val=Value(val.getValue(), currency);
      account().ref().setCreditLine(val);
    }



  }
  catch (Error xerr) {
    return Error("JobGetBalance::evaluate", xerr);
  }
  return Error();
}





JobGetTransactions::JobGetTransactions(Pointer<Customer> c,
                                       Pointer<Account> acc,
                                       const Date &fromDate,
                                       const Date &toDate)
:AccountJob("JOBGetTransactions", c, acc) {
  if (fromDate.isValid())
    setProperty("fromdate", fromDate.toString());
  if (toDate.isValid())
    setProperty("todate", toDate.toString());
}



JobGetTransactions::~JobGetTransactions(){
}


Error JobGetTransactions::_parseMT(const string &mt,
                                   GWEN_DB_NODE *rsp,
                                   const string &grpname) {
  Error err;
  GWEN_BUFFER *tbuf;
  GWEN_BUFFEREDIO *bio;
  Pointer<AQMConverter> conv;
  GWEN_DB_NODE *transactions;

  if (0) {
    FILE *f;
    static unsigned int count=0;
    char buffer[128];

    snprintf(buffer, sizeof(buffer)-1, "/tmp/mt94x-%03d.mt", count++);
    buffer[sizeof(buffer)-1]=0;
    f=fopen(buffer, "w+");
    if (f){
      fwrite(mt.data(), mt.length(), 1, f);
      fclose(f);
    }
  }

  tbuf=GWEN_Buffer_new(0, mt.length(), 0, 1);
  GWEN_Buffer_AppendBytes(tbuf, mt.data(), mt.length());

  /* data ready */
  GWEN_Buffer_Rewind(tbuf);
  bio=GWEN_BufferedIO_Buffer_new(tbuf);
  GWEN_BufferedIO_SetReadBuffer(bio, 0, 128);

  // create converter
  conv=api()->converterFactory("SWIFT", "mt940");
  if (!conv.isValid()) {
    DBG_ERROR(0, "No SWIFT parser ?");
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    return Error("JobGetTransactions::commit",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "could not create converter for SWIFT");
  }
  /* read transactions in */
  err=conv.ref().readFromStream(bio);
  if (!err.isOk()) {
    DBG_INFO(0, "Error reading SWIFT document: %s\n",
	     err.errorString().c_str());
    GWEN_BufferedIO_Abandon(bio);
    GWEN_BufferedIO_free(bio);
    return Error("JobGetTransactions::commit", err);
  }
  transactions=conv.ref().transactions();
  if (transactions) {
    GWEN_DB_NODE *dstgrp;

    DBG_INFO(0, "I have these transactions:");
    if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelInfo)
      GWEN_DB_Dump(transactions, stderr, 2);
    dstgrp=GWEN_DB_GetGroup(rsp,
                            GWEN_DB_FLAGS_DEFAULT,
                            grpname.c_str());
    assert(dstgrp);
    GWEN_DB_AddGroupChildren(dstgrp, transactions);
  }

  GWEN_BufferedIO_Close(bio);
  GWEN_BufferedIO_free(bio);

  return Error();
}



Error JobGetTransactions::commit(){
  string currency;
  Value val;
  GWEN_DB_NODE *rsp;
  GWEN_DB_NODE *reportn;
  Pointer<TransactionStorage> st;
  string currentDay;
  unsigned int dayhdl;
  Error err;
  DateTime today;
  char currentDate[9];

  today=DateTime::currentTime();
  snprintf(currentDate, sizeof(currentDate), "%04d%02d%02d",
           today.year(), today.month(), today.day());

  rsp=responseData();
  if (!rsp)
    return Error("JobGetTransactions::commit",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no results");
  rsp=GWEN_DB_GetGroup(rsp, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "transactions");
  if (!rsp) {
    DBG_INFO(0, "No entries");
    return Error();
  }

  reportn=GWEN_DB_GetGroup(rsp,
			   GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			   "booked");
  if (!reportn) {
    GWEN_DB_NODE *tmp;
    string mt;

    /* we have no entries yet, so we might need to parse binary data
     * ourself*/
    tmp=GWEN_DB_GetFirstGroup(responseData());
    while(tmp) {
      if (strcasecmp(GWEN_DB_GroupName(tmp), "transactions")==0) {
        const void *p;
        unsigned int bs;

        DBG_INFO(0, "No entries, trying to parse binary data");
        p=GWEN_DB_GetBinValue(tmp, "booked", 0, 0, 0, &bs);
        if (p && bs) {
          mt+=string((const char*)p, bs);
        }
      } /* if transactions */
      tmp=GWEN_DB_GetNextGroup(tmp);
    } /* while */

    err=_parseMT(mt, rsp, "booked");
    if (!err.isOk()) {
      return Error("JobGetTransactions::commit", err);
    }
  }

  reportn=GWEN_DB_GetGroup(rsp,
			   GWEN_PATH_FLAGS_NAMEMUSTEXIST,
			   "noted");
  if (!reportn) {
    GWEN_DB_NODE *tmp;
    string mt;

    /* we have no entries yet, so we might need to parse binary data
     * ourself*/
    tmp=GWEN_DB_GetFirstGroup(responseData());
    while(tmp) {
      if (strcasecmp(GWEN_DB_GroupName(tmp), "transactions")==0) {
        const void *p;
        unsigned int bs;

        DBG_INFO(0, "No entries, trying to parse binary data");
        p=GWEN_DB_GetBinValue(tmp, "noted", 0, 0, 0, &bs);
        if (p && bs) {
          mt+=string((const char*)p, bs);
        }
      } /* if transactions */
      tmp=GWEN_DB_GetNextGroup(tmp);
    } /* while */

    err=_parseMT(mt, rsp, "noted");
    if (!err.isOk()) {
      return Error("JobGetTransactions::commit", err);
    }
  }

  DBG_DEBUG(0, "This is my result:\n");
  if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
    GWEN_DB_Dump(rsp, stderr, 1);
  try {
    st=account().ref().storage();
    reportn=GWEN_DB_GetGroup(rsp,
                             GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                             "booked");
    if (reportn) {
      reportn=GWEN_DB_GetFirstGroup(reportn);
      dayhdl=0;
      currentDay="";
      while(reportn) {
        if (strcasecmp(GWEN_DB_GroupName(reportn), "report")==0) {
          GWEN_DB_NODE *cn;

          cn=GWEN_DB_GetFirstGroup(reportn);
          while(cn) {
            if (strcasecmp(GWEN_DB_GroupName(cn), "transaction")==0) {
              const char *p;
              Date d;

              DBG_INFO(0, "Found transaction");
              /* store current date for Florian */
              GWEN_DB_SetCharValue(cn, GWEN_DB_FLAGS_OVERWRITE_VARS,
                                   "storedate",
                                   currentDate);
              p=GWEN_DB_GetCharValue(cn, "date", 0, 0);
              if (!p)
                p=GWEN_DB_GetCharValue(cn, "valutaDate", 0, 0);
              if (p) {
                // we have a date, so try to open that day
                if (strcasecmp(p, currentDay.c_str())!=0) {
                  // new day, so close the old one, if needed
                  if (dayhdl) {
                    DBG_INFO(0, "Closing day \"%s\"",
                             currentDay.c_str());
                    err=st.ref().closeDay(dayhdl);
                    if (!err.isOk())
                      return Error("JobGetTransactions::commit", err);
                    DBG_INFO(0, "Closing day \"%s\" done.",
                             currentDay.c_str());
                  } // if day open
                  // open new day
                  d=Date(p);
                  DBG_INFO(0, "Opening day \"%s\"", p);
                  dayhdl=st.ref().openDay(d.year(),
                                          d.month(),
                                          d.day(),
                                          true);
                  DBG_INFO(0, "Day \"%s\" is open", p);
                  err=st.ref().clearDay(dayhdl);
                  if (!err.isOk())
                    return Error("JobGetTransactions::commit", err);
                  currentDay=p;
                } // if new day
                DBG_INFO(0, "Adding transaction");
                err=st.ref().addTransaction(dayhdl, cn);
                if (!err.isOk())
                  return Error("JobGetTransactions::commit", err);
                DBG_INFO(0, "Adding transaction done.");

              } // if date given
            }
            else if (strcasecmp(GWEN_DB_GroupName(cn), "startsaldo")==0) {
              DBG_INFO(0, "Found opening balance");
            }
            else if (strcasecmp(GWEN_DB_GroupName(cn), "endsaldo")==0) {
              DBG_INFO(0, "Found closing balance");
            }
            else {
              DBG_WARN(0, "Unknown section \"%s\"", GWEN_DB_GroupName(cn));
            }
            cn=GWEN_DB_GetNextGroup(cn);
          } // while cn
        } // if report
        else {
          DBG_WARN(0, "Unknown section \"%s\"", GWEN_DB_GroupName(reportn));
        }
        reportn=GWEN_DB_GetNextGroup(reportn);
      } // while reports
      if (dayhdl) {
        DBG_INFO(0, "Closing day \"%s\"",
                 currentDay.c_str());
        err=st.ref().closeDay(dayhdl);
        if (!err.isOk())
          return Error("JobGetTransactions::commit", err);
        DBG_INFO(0, "Closing day \"%s\" done.",
                 currentDay.c_str());
      } // if day open
    } // if booked
    else {
      DBG_WARN(0, "No \"booked\" section");
    }

    // clear noted balance, because we are about to import new ones ( if
    // there are any)
    account().ref().clearNotedTransfers();
    // get noted transfers
    reportn=GWEN_DB_GetGroup(rsp,
                             GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                             "noted");
    if (reportn) {
      reportn=GWEN_DB_GetFirstGroup(reportn);
      while(reportn) {
        if (strcasecmp(GWEN_DB_GroupName(reportn), "report")==0) {
          GWEN_DB_NODE *cn;

          cn=GWEN_DB_GetFirstGroup(reportn);
          while(cn) {
            if (strcasecmp(GWEN_DB_GroupName(cn), "transaction")==0) {
              // add noted transfer
              err=account().ref().addNotedTransfer(cn);
              if (!err.isOk()) {
                return Error("JobGetTransactions::commit", err);
              }
            }
            else if (strcasecmp(GWEN_DB_GroupName(cn), "startsaldo")==0) {
              DBG_INFO(0, "Found opening balance");
            }
            else if (strcasecmp(GWEN_DB_GroupName(cn), "endsaldo")==0) {
              DBG_INFO(0, "Found closing balance");
            }
            else {
              DBG_WARN(0, "Unknown section \"%s\"", GWEN_DB_GroupName(cn));
            }
            cn=GWEN_DB_GetNextGroup(cn);
          } // while cn
        } // if report
        else {
          DBG_WARN(0, "Unknown section \"%s\"", GWEN_DB_GroupName(reportn));
        }
        reportn=GWEN_DB_GetNextGroup(reportn);
      } // while reports
    } // if reports

  } // try
  catch (Error xerr) {
    return Error("JobGetTransactions::commit", xerr);
  }
  return Error();
}





JobGetStandingOrders::JobGetStandingOrders(Pointer<Customer> c,
                                           Pointer<Account> acc)
:AccountJob("JOBGetStandingOrders", c, acc) {
}



JobGetStandingOrders::~JobGetStandingOrders(){
}



Error JobGetStandingOrders::commit(){
  GWEN_DB_NODE *rsp;
  GWEN_DB_NODE *n;
  Error err;
  Pointer<Account> acc;

  acc=account();
  rsp=responseData();
  if (!rsp)
    return Error("JobGetStandingOrders::commit",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no results");
  acc.ref().clearStandingOrders();
  n=GWEN_DB_GetFirstGroup(rsp);
  while(n) {
    if (strcasecmp(GWEN_DB_GroupName(n), "GetStandingOrdersResponse")==0) {
      GWEN_DB_NODE *nn;

      nn=GWEN_DB_GetGroup(n,
                          GWEN_DB_FLAGS_DEFAULT |
                          GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                          "standingorder");
      if (nn)
        acc.ref().addStandingOrder(nn);
      else {
        DBG_WARN(0, "No standing order in response...");
      }
    }
    n=GWEN_DB_GetNextGroup(n);
  } // while

  return Error();
}








JobSingleTransferBase::JobSingleTransferBase(const string &name,
                                             Pointer<Customer> c,
                                             Pointer<Account> acc,
                                             GWEN_DB_NODE *xa)
:AccountJob(name, c, acc)
,_transaction(0) {
  Error err;

  err=setTransaction(xa);
  if (!err.isOk())
    throw Error("JobSingleTransferBase::JobSingleTransferBase", err);
}



JobSingleTransferBase::JobSingleTransferBase(const string &name,
                                             Pointer<Customer> c,
                                             Pointer<Account> acc)
:AccountJob(name, c, acc)
,_transaction(0) {
}



JobSingleTransferBase::~JobSingleTransferBase(){
  GWEN_DB_Group_free(_transaction);
}



Error JobSingleTransferBase::evaluate(Pointer<Outbox> outbox){
  Error err;
  list<MsgJob::Result>::const_iterator rit;
  GWEN_DB_NODE *rn;
  GWEN_DB_NODE *n;
  string s;

  err=AccountJob::evaluate(outbox);

  if (status()==HBCI_JOB_STATUS_DONE) {
    // store global results of this job
    rn=GWEN_DB_GetGroup(_transaction,
                        GWEN_DB_FLAGS_DEFAULT,
                        "results");
    assert(rn);

    // store status
    switch(status()) {
    case HBCI_JOB_STATUS_TODO: s="todo"; break;
    case HBCI_JOB_STATUS_DONE: s="done"; break;
    default:                   s="unknown"; break;
    }
    GWEN_DB_SetCharValue(rn, GWEN_DB_FLAGS_DEFAULT,
                         "statusCode",
                         s.c_str());

    // store result
    switch(result()) {
    case HBCI_JOB_RESULT_SUCCESS:        s="success"; break;
    case HBCI_JOB_RESULT_PARTLY_SUCCESS: s="partly"; break;
    case HBCI_JOB_RESULT_FAILED:         s="failed"; break;
    default:                             s="unknown"; break;
    }
    GWEN_DB_SetCharValue(rn, GWEN_DB_FLAGS_DEFAULT,
                         "resultCode",
                         s.c_str());

    // store each job result
    for (rit=results().begin();
         rit!=results().end();
         rit++) {
      n=GWEN_DB_GetGroup(rn,
                         GWEN_DB_FLAGS_DEFAULT | GWEN_PATH_FLAGS_CREATE_GROUP,
                         "jobresult");
      GWEN_DB_SetIntValue(n, GWEN_DB_FLAGS_DEFAULT,
                          "resultCode",
                          (*rit).resultCode());
      GWEN_DB_SetCharValue(n, GWEN_DB_FLAGS_DEFAULT,
                           "resultText",
                           (*rit).resultMessage().c_str());
    } // for
  } // if done

  return Error("JobSingleTransferBase::evaluate", err);
}



Error JobSingleTransferBase::commit(){
  Pointer<TransactionStorage> st;
  unsigned int dayhdl;
  Error err;
  string p;
  Date d;
  DateTime dt;

  // get current time
  dt=DateTime::currentTime();
  p=dt.timeString("%Y%N%D");

  // store transfer
  try {
    st=account().ref().transferStorage();
    // open new day
    d=Date(p);
    DBG_INFO(0, "Opening day \"%s\"", p.c_str());
    dayhdl=st.ref().openDay(d.year(),
                            d.month(),
                            d.day(),
                            true);
    DBG_INFO(0, "Day \"%s\" is open", p.c_str());

    DBG_INFO(0, "Adding transaction");
    if (status()!=HBCI_JOB_STATUS_DONE) {
      string s;

      // store status
      switch(status()) {
      case HBCI_JOB_STATUS_TODO: s="todo";   break;
      case HBCI_JOB_STATUS_DONE: s="done";   break;
      default:                   s="unsent"; break;
      }
      GWEN_DB_SetCharValue(_transaction, GWEN_DB_FLAGS_DEFAULT,
                           "results/statusCode",
                           s.c_str());
    }
    err=st.ref().addTransaction(dayhdl, _transaction);
    if (!err.isOk())
      return Error("JobSingleTransferBase::commit", err);
    DBG_INFO(0, "Adding transaction done.");

    DBG_INFO(0, "Closing day \"%s\"", p.c_str());
    err=st.ref().closeDay(dayhdl);
    if (!err.isOk())
      return Error("JobGetTransactions::commit", err);
    DBG_INFO(0, "Closing day \"%s\" done.", p.c_str());
  } // try
  catch (Error xerr) {
    return Error("JobSingleTransferBase::commit", xerr);
  }
  return Error();
}



Error JobSingleTransferBase::setTransaction(GWEN_DB_NODE *xa){
  Value val;

  if (_transaction)
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "transaction already set");

  try {
    val=Value(GWEN_DB_GetCharValue(xa, "value/value", 0,"0,"),
              GWEN_DB_GetCharValue(xa, "value/currency", 0,""));
  }
  catch (Error xerr) {
    return Error("JobSingleTransfer::setTransaction", xerr);
  }

  // check value
  if (val.getValue()==0.0) {
    DBG_ERROR(0, "Zero not allowed as value for transfers");
    AccountJob::storeResult(xa, 9999,
                            "zero value not allowed");
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "zero value not allowed");
  }
  if (val.getCurrency().empty()) {
    DBG_ERROR(0, "Currency needed for transfers");
    AccountJob::storeResult(xa, 9999,
                            "currency needed");
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "currency needed");
  }

  _transaction=GWEN_DB_Group_dup(xa);
  addRequestData("", _transaction);
  return Error();
}






JobSingleTransfer::JobSingleTransfer(Pointer<Customer> c,
                                     Pointer<Account> acc,
                                     GWEN_DB_NODE *transaction)
:JobSingleTransferBase("JobSingleTransfer", c, acc,
                       transaction) {
}



JobSingleTransfer::JobSingleTransfer(Pointer<Customer> c,
                                     Pointer<Account> acc)
:JobSingleTransferBase("JobSingleTransfer", c, acc) {
}



JobSingleTransfer::~JobSingleTransfer(){
}



Error JobSingleTransfer::commit(){
  return JobSingleTransferBase::commit();
}



JobSingleDebitNote::JobSingleDebitNote(Pointer<Customer> c,
                                     Pointer<Account> acc,
                                     GWEN_DB_NODE *transaction)
:JobSingleTransferBase("JobSingleDebitNote", c, acc,
                       GWEN_DB_Group_dup(transaction)) {
}



JobSingleDebitNote::JobSingleDebitNote(Pointer<Customer> c,
                                     Pointer<Account> acc)
:JobSingleTransferBase("JobSingleDebitNote", c, acc) {
}



JobSingleDebitNote::~JobSingleDebitNote(){
}



Error JobSingleDebitNote::commit(){
  return JobSingleTransferBase::commit();
}






/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobDatedSingleTransfer
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */

JobDatedSingleTransfer::JobDatedSingleTransfer(Pointer<Customer> c,
                                               Pointer<Account> acc,
                                               GWEN_DB_NODE *transaction)
:JobSingleTransferBase("JobDatedSingleTransfer", c, acc, transaction) {

}



JobDatedSingleTransfer::JobDatedSingleTransfer(Pointer<Customer> c,
                                               Pointer<Account> acc)
:JobSingleTransferBase("JobDatedSingleTransfer", c, acc) {

}


Error JobDatedSingleTransfer::evaluate(Pointer<Outbox> outbox){
  Error err;
  GWEN_DB_NODE *rsp;

  err=JobSingleTransferBase::evaluate(outbox);
  if (!err.isOk())
    return Error("JobDatedSingleTransfer::evaluate", err);

  rsp=responseData();
  if (!rsp)
    return Error("JobDatedSingleTransfer::evaluate",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no results");
  rsp=GWEN_DB_GetGroup(rsp,
                       GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                       "DatedSingleTransferResponse");
  if (!rsp)
    return Error("JobDatedSingleTransfer::evaluate",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no result for this job, maybe an error occurred ?");
  _jobId=GWEN_DB_GetCharValue(rsp, "referenceId", 0, "");
  DBG_ERROR(0, "Job id is: %s", _jobId.c_str());

  return Error();
}




JobDatedSingleTransfer::~JobDatedSingleTransfer(){
}



Error JobDatedSingleTransfer::commit(){
  Error err;

  /* store jobid */
  if (!_jobId.empty())
    GWEN_DB_SetCharValue(transaction(),
                         GWEN_DB_FLAGS_DEFAULT |
                         GWEN_DB_FLAGS_OVERWRITE_VARS,
                         "jobid", _jobId.c_str());
  err=JobSingleTransferBase::commit();
  return Error("JobDatedSingleTransfer::commit", err);
}




/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobChangeDatedSingleTransfer
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */

JobChangeDatedSingleTransfer::JobChangeDatedSingleTransfer(Pointer<Customer> c,
                                               Pointer<Account> acc,
                                               GWEN_DB_NODE *xa)
:AccountJob("JobChangeDatedSingleTransfer", c, acc)
,_transfer(0) {
  Error err;

  err=setTransfer(xa);
  if (!err.isOk())
    throw Error("JobChangeDatedSingleTransfer::JobChangeDatedSingleTransfer",
                err);
}



JobChangeDatedSingleTransfer::JobChangeDatedSingleTransfer(Pointer<Customer> c,
                                               Pointer<Account> acc)
:AccountJob("JobChangeDatedSingleTransfer", c, acc)
,_transfer(0) {

}



Error JobChangeDatedSingleTransfer::evaluate(Pointer<Outbox> outbox){
  Error err;

  err=AccountJob::evaluate(outbox);
  if (!err.isOk())
    return Error("JobChangeDatedSingleTransfer::evaluate", err);

  return Error();
}



Error JobChangeDatedSingleTransfer::commit(){
  return Error();
}



Error JobChangeDatedSingleTransfer::setTransfer(GWEN_DB_NODE *xa){
  Value val;

  if (_transfer)
    return Error("JobChangeDatedSingleTransfer::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "transaction already set");

  try {
    val=Value(GWEN_DB_GetCharValue(xa, "value/value", 0,"0,"),
              GWEN_DB_GetCharValue(xa, "value/currency", 0,""));
  }
  catch (Error xerr) {
    return Error("JobSingleTransfer::setTransaction", xerr);
  }

  // check value
  if (val.getValue()==0.0) {
    DBG_ERROR(0, "Zero not allowed as value for transfers");
    AccountJob::storeResult(xa, 9999,
                            "zero value not allowed");
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "zero value not allowed");
  }
  if (val.getCurrency().empty()) {
    DBG_ERROR(0, "Currency needed for transfers");
    AccountJob::storeResult(xa, 9999,
                            "currency needed");
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "currency needed");
  }

  _transfer=GWEN_DB_Group_dup(xa);
  addRequestData("", _transfer);
  return Error();
}



JobChangeDatedSingleTransfer::~JobChangeDatedSingleTransfer(){
}





/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobDeleteDatedSingleTransfer
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */

JobDeleteDatedSingleTransfer::JobDeleteDatedSingleTransfer(Pointer<Customer> c,
                                               Pointer<Account> acc,
                                               GWEN_DB_NODE *xa)
:AccountJob("JobDeleteDatedSingleTransfer", c, acc)
,_transfer(0) {
  Error err;

  err=setTransfer(xa);
  if (!err.isOk())
    throw Error("JobDeleteDatedSingleTransfer::JobDeleteDatedSingleTransfer",
                err);
}



JobDeleteDatedSingleTransfer::JobDeleteDatedSingleTransfer(Pointer<Customer> c,
                                               Pointer<Account> acc)
:AccountJob("JobDeleteDatedSingleTransfer", c, acc)
,_transfer(0) {

}



Error JobDeleteDatedSingleTransfer::evaluate(Pointer<Outbox> outbox){
  Error err;

  err=AccountJob::evaluate(outbox);
  if (!err.isOk())
    return Error("JobDeleteDatedSingleTransfer::evaluate", err);

  return Error();
}



Error JobDeleteDatedSingleTransfer::commit(){
  return Error();
}



Error JobDeleteDatedSingleTransfer::setTransfer(GWEN_DB_NODE *xa){
  Value val;

  if (_transfer)
    return Error("JobDeleteDatedSingleTransfer::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "transaction already set");

  try {
    val=Value(GWEN_DB_GetCharValue(xa, "value/value", 0,"0,"),
              GWEN_DB_GetCharValue(xa, "value/currency", 0,""));
  }
  catch (Error xerr) {
    return Error("JobSingleTransfer::setTransaction", xerr);
  }

  // check value
  if (val.getValue()==0.0) {
    DBG_ERROR(0, "Zero not allowed as value for transfers");
    AccountJob::storeResult(xa, 9999,
                            "zero value not allowed");
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "zero value not allowed");
  }
  if (val.getCurrency().empty()) {
    DBG_ERROR(0, "Currency needed for transfers");
    AccountJob::storeResult(xa, 9999,
                            "currency needed");
    return Error("JobSingleTransferBase::setTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "currency needed");
  }

  _transfer=GWEN_DB_Group_dup(xa);
  addRequestData("", _transfer);
  return Error();
}



JobDeleteDatedSingleTransfer::~JobDeleteDatedSingleTransfer(){
}




/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobGetDatedTransfers
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */

JobGetDatedTransfers::JobGetDatedTransfers(Pointer<Customer> c,
                                           Pointer<Account> acc,
                                           const Date &fromDate,
                                           const Date &toDate)
:AccountJob("JobGetDatedSingleTransfers", c, acc) {
  if (fromDate.isValid())
    setProperty("fromdate", fromDate.toString());
  if (toDate.isValid())
    setProperty("todate", toDate.toString());
}



Error JobGetDatedTransfers::evaluate(Pointer<Outbox> outbox){
  Error err;

  err=AccountJob::evaluate(outbox);
  if (!err.isOk())
    return Error("JobGetDatedTransfers::evaluate", err);

  return Error();
}




JobGetDatedTransfers::~JobGetDatedTransfers(){
}



Error JobGetDatedTransfers::commit(){
  string currency;
  Value val;
  GWEN_DB_NODE *rsp;
  GWEN_DB_NODE *reportn;
  Error err;
  Pointer<Account> acc;

  rsp=responseData();
  if (!rsp)
    return Error("JobGetDatedTransfers::commit",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "no results");
  acc=account();
  DBG_DEBUG(0, "This is my result:\n");
  if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
    GWEN_DB_Dump(rsp, stderr, 1);
  try {
    acc.ref().clearDatedTransfers();
    reportn=GWEN_DB_GetFirstGroup(rsp);
    while(reportn) {
      if (strcasecmp(GWEN_DB_GroupName(reportn),
                     "GetDatedTransfersResponse")==0) {
        GWEN_DB_NODE *nn;

        nn=GWEN_DB_GetGroup(reportn,
                            GWEN_DB_FLAGS_DEFAULT |
                            GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                            "datedtransfer");
        if (nn) {
          DBG_INFO(0, "Adding transaction");
          err=acc.ref().addDatedTransfer(nn);
          if (!err.isOk())
            return Error("JobGetDatedTransfers::commit", err);
        } // if transaction
        else {
          DBG_WARN(0, "No dated transfer in response...");
        }
      }
      reportn=GWEN_DB_GetNextGroup(reportn);
    } // while reports
  } // try
  catch (Error xerr) {
    return Error("JobGetDatedTransfers::commit", xerr);
  }
  return Error();
}







JobMultiTransferBase::JobMultiTransferBase(const string &name,
                                           const string &defTextKey,
                                           Pointer<Customer> c,
                                           Pointer<Account> acc)
:AccountJob(name, c, acc)
,_transactions(GWEN_DB_Group_new("transactions"))
,_defTextKey(defTextKey)
,_transactionCount(0){
}



Error JobMultiTransferBase::addTransaction(GWEN_DB_NODE *xa) {
  Value val;
  GWEN_DB_NODE *params;
  int idtk;
  const char *p;

  idtk=atoi(_defTextKey.c_str());

  try {
    val=Value(GWEN_DB_GetCharValue(xa, "value/value", 0,"0,"),
              GWEN_DB_GetCharValue(xa, "value/currency", 0,""));
  }
  catch (Error xerr) {
    return Error("JobSingleTransfer::JobSingleTransfer", xerr);
  }

  // check value
  if (val.getValue()==0.0) {
    DBG_ERROR(0, "Zero not allowed as value for transfers");
    return Error("JobMultiTransferBase::addTransAction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "zero value not allowed");
  }
  if (val.getCurrency().empty()) {
    DBG_ERROR(0, "Currency needed for transfers");
    return Error("JobMultiTransferBase::addTransAction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "currency needed");
  }

  // check textkey
  params=paramData();
  if (params) {
    GWEN_DB_NODE *n;
    int tk;

    tk=GWEN_DB_GetIntValue(xa, "textkey", 0, -1);
    if (tk==-1)
      tk=atoi(GWEN_DB_GetCharValue(xa, "textkey", 0, "-1"));
    if (tk==-1) {
      GWEN_DB_SetCharValue(xa, GWEN_DB_FLAGS_DEFAULT,
                           "textkey", _defTextKey.c_str());
      tk=atoi(_defTextKey.c_str());
    }

    // check textkey
    n=GWEN_DB_GetGroup(params, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "textKey");
    if (n) {
      int i;

      for (i=0;;i++) {
        int ck;

        ck=GWEN_DB_GetIntValue(n, "key", i, -1);
        if (ck==-1) {
          if (i) {
            return Error("JobMultiTransferBase::addTransAction",
                         ERROR_LEVEL_NORMAL,
                         HBCI_ERROR_CODE_INVALID,
                         ERROR_ADVISE_DONTKNOW,
                         "textkey not supported for this account",
                         String::num2string(tk));
          }
          break;
        }
        if (ck==tk)
          break;
      } // for
    } // if textKey given
  } // if params found

  p=GWEN_DB_GetCharValue(xa, "date", 0, 0);
  if (p) {
    DBG_INFO(0, "date not allowed for multi transfers");
    return Error("JobMultiTransferBase::addTransAction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "date not allowed for multi transfers");
  }

  p=GWEN_DB_GetCharValue(xa, "name", 0, 0);
  if (!p)
    GWEN_DB_SetCharValue(xa, GWEN_DB_FLAGS_DEFAULT, "name",
                         account().ref().userName().c_str());

  GWEN_DB_AddGroup(_transactions, GWEN_DB_Group_dup(xa));
  if (idtk==4 || idtk==5)
    addRequestData("transfers/debitnote", xa);
  else
    addRequestData("transfers/transaction", xa);
  _transactionCount++;
  return Error();
}



JobMultiTransferBase::~JobMultiTransferBase(){
}



Error JobMultiTransferBase::evaluate(Pointer<Outbox> outbox){
  Error err;
  list<MsgJob::Result>::const_iterator rit;
  GWEN_DB_NODE *rn;
  GWEN_DB_NODE *n;
  string s;

  err=AccountJob::evaluate(outbox);

  if (status()==HBCI_JOB_STATUS_DONE) {
    GWEN_DB_NODE *tr;

    tr=GWEN_DB_GetFirstGroup(_transactions);
    while (tr) {
      // store global results of this job
      rn=GWEN_DB_GetGroup(tr,
                          GWEN_DB_FLAGS_DEFAULT,
                          "results");
      assert(rn);
  
      // store status
      switch(status()) {
      case HBCI_JOB_STATUS_TODO: s="todo"; break;
      case HBCI_JOB_STATUS_DONE: s="done"; break;
      default:                   s="unknown"; break;
      }
      GWEN_DB_SetCharValue(rn, GWEN_DB_FLAGS_DEFAULT,
                           "statusCode",
                           s.c_str());
  
      // store result
      switch(result()) {
      case HBCI_JOB_RESULT_SUCCESS:        s="success"; break;
      case HBCI_JOB_RESULT_PARTLY_SUCCESS: s="partly"; break;
      case HBCI_JOB_RESULT_FAILED:         s="failed"; break;
      default:                             s="unknown"; break;
      }
      GWEN_DB_SetCharValue(rn, GWEN_DB_FLAGS_DEFAULT,
                           "resultCode",
                           s.c_str());
  
      // store each job result
      for (rit=results().begin();
           rit!=results().end();
           rit++) {
        n=GWEN_DB_GetGroup(rn,
                           GWEN_DB_FLAGS_DEFAULT |
                           GWEN_PATH_FLAGS_CREATE_GROUP,
                           "jobresult");
        GWEN_DB_SetIntValue(n, GWEN_DB_FLAGS_DEFAULT,
                            "resultCode",
                            (*rit).resultCode());
        GWEN_DB_SetCharValue(n, GWEN_DB_FLAGS_DEFAULT,
                             "resultText",
                             (*rit).resultMessage().c_str());
      } // for
      tr=GWEN_DB_GetNextGroup(tr);
    } // while transfers
  } // if done

  return Error("JobMultiTransferBase::evaluate", err);
}



Error JobMultiTransferBase::commit(){
  Pointer<TransactionStorage> st;
  unsigned int dayhdl;
  Error err;
  Error lerr;
  string p;
  Date d;
  DateTime dt;
  GWEN_DB_NODE *tr;

  // get current time
  dt=DateTime::currentTime();
  p=dt.timeString("%Y%N%D");
  try {
    d=Date(p);
    st=account().ref().transferStorage();
    DBG_INFO(0, "Opening day \"%s\"", p.c_str());
    // open new day
    dayhdl=st.ref().openDay(d.year(),
                            d.month(),
                            d.day(),
                            true);
    DBG_INFO(0, "Day \"%s\" is open", p.c_str());
  }
  catch (Error xerr) {
    return Error("JobMultiTransferBase::commit", xerr);
  }

  // store transfers
  tr=GWEN_DB_GetFirstGroup(_transactions);
  while(tr) {
    DBG_INFO(0, "Adding transaction");
    if (status()!=HBCI_JOB_STATUS_DONE) {
      string s;

      // store status
      switch(status()) {
      case HBCI_JOB_STATUS_TODO: s="todo";   break;
      case HBCI_JOB_STATUS_DONE: s="done";   break;
      default:                   s="unsent"; break;
      }
      GWEN_DB_SetCharValue(tr, GWEN_DB_FLAGS_DEFAULT,
                           "results/statusCode",
                           s.c_str());
    }
    lerr=st.ref().addTransaction(dayhdl, tr);
    if (!lerr.isOk())
      err=lerr;
    DBG_INFO(0, "Adding transaction done.");
    tr=GWEN_DB_GetNextGroup(tr);
  } // while tr

  DBG_INFO(0, "Closing day \"%s\"", p.c_str());
  lerr=st.ref().closeDay(dayhdl);
  if (!lerr.isOk())
    err=lerr;
  DBG_INFO(0, "Closing day \"%s\" done.", p.c_str());

  return err;
}






JobMultiTransfer::JobMultiTransfer(Pointer<Customer> c,
                                   Pointer<Account> acc)
:JobMultiTransferBase("JobMultiTransfer", "51", c, acc) {
}



JobMultiTransfer::~JobMultiTransfer(){
}



Error JobMultiTransfer::commit(){
  return JobMultiTransferBase::commit();
}




JobMultiDebitNote::JobMultiDebitNote(Pointer<Customer> c,
                                     Pointer<Account> acc)
:JobMultiTransferBase("JobMultiDebitNote", "5", c, acc) {
}



JobMultiDebitNote::~JobMultiDebitNote(){
}



Error JobMultiDebitNote::commit(){
  return JobMultiTransferBase::commit();
}





/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobStandingOrderBase
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */



JobStandingOrderBase::JobStandingOrderBase(const string &name,
                                           Pointer<Customer> c,
                                           Pointer<Account> acc,
                                           GWEN_DB_NODE *xa)
:AccountJob(name, c ,acc)
,_sto(0) {
  Error err;

  err=setStandingOrder(xa);
  if (!err.isOk())
    throw Error("JobStandingOrderBase::JobStandingOrderBase", err);
}



JobStandingOrderBase::JobStandingOrderBase(const string &name,
                                           Pointer<Customer> c,
                                           Pointer<Account> acc)
:AccountJob(name, c ,acc)
,_sto(0) {
}



JobStandingOrderBase::~JobStandingOrderBase(){
  GWEN_DB_Group_free(_sto);
}


Error JobStandingOrderBase::commit(){
  return Error();
}



Error JobStandingOrderBase::evaluate(Pointer<Outbox> outbox){
  Error err;
  list<MsgJob::Result>::const_iterator rit;
  GWEN_DB_NODE *rn;
  GWEN_DB_NODE *n;
  string s;

  err=AccountJob::evaluate(outbox);

  if (status()==HBCI_JOB_STATUS_DONE) {
    // store global results of this job
    rn=GWEN_DB_GetGroup(_sto,
                        GWEN_DB_FLAGS_DEFAULT,
                        "results");
    assert(rn);

    // store status
    switch(status()) {
    case HBCI_JOB_STATUS_TODO: s="todo"; break;
    case HBCI_JOB_STATUS_DONE: s="done"; break;
    default:                   s="unknown"; break;
    }
    GWEN_DB_SetCharValue(rn, GWEN_DB_FLAGS_DEFAULT,
                         "statusCode",
                         s.c_str());

    // store result
    switch(result()) {
    case HBCI_JOB_RESULT_SUCCESS:        s="success"; break;
    case HBCI_JOB_RESULT_PARTLY_SUCCESS: s="partly"; break;
    case HBCI_JOB_RESULT_FAILED:         s="failed"; break;
    default:                             s="unknown"; break;
    }
    GWEN_DB_SetCharValue(rn, GWEN_DB_FLAGS_DEFAULT,
                         "resultCode",
                         s.c_str());

    // store each job result
    for (rit=results().begin();
         rit!=results().end();
         rit++) {
      n=GWEN_DB_GetGroup(rn,
                         GWEN_DB_FLAGS_DEFAULT | GWEN_PATH_FLAGS_CREATE_GROUP,
                         "jobresult");
      GWEN_DB_SetIntValue(n, GWEN_DB_FLAGS_DEFAULT,
                          "resultCode",
                          (*rit).resultCode());
      GWEN_DB_SetCharValue(n, GWEN_DB_FLAGS_DEFAULT,
                           "resultText",
                           (*rit).resultMessage().c_str());
    } // for
  } // if done

  return Error("JobStandingOrderBase::evaluate", err);
}


Error JobStandingOrderBase::setStandingOrder(GWEN_DB_NODE *xa){
  Value val;
  int i, j;
  GWEN_DB_NODE *params;
  const char *p;
  unsigned int k;
  string tstr;
  string dstr;
  bool supported;

  if (_sto)
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "standing order already set");

  // check value
  try {
    val=Value(GWEN_DB_GetCharValue(xa, "value/value", 0,"0,"),
              GWEN_DB_GetCharValue(xa, "value/currency", 0,""));
  }
  catch (Error xerr) {
    return Error("JobStandingOrderBase::setStandingOrder", xerr);
  }
  if (val.getValue()==0.0) {
    DBG_ERROR(0, "Zero not allowed as value for transfers");
    AccountJob::storeResult(xa, 9999,
                            "zero value not allowed");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "zero value not allowed");
  }
  if (val.getCurrency().empty()) {
    DBG_ERROR(0, "Currency needed for transfers");
    AccountJob::storeResult(xa, 9999,
                            "currency needed");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "currency needed");
  }

  // check unit, turnus and day
  p=GWEN_DB_GetCharValue(xa, "unit", 0, 0);
  if (!p) {
    DBG_ERROR(0, "unit needed, either W or M");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "unit needed, either W or M");
  }
  if (toupper(*p)=='W') {
    tstr=GWEN_DB_GetCharValue(paramData(), "AllowedTurnusWeeks", 0, "00");
    dstr=GWEN_DB_GetCharValue(paramData(), "AllowedWeekDays", 0, "0");

    // check day
    i=GWEN_DB_GetIntValue(xa, "day", 0, -1);
    if (i==-1) {
      DBG_ERROR(0, "day needed");
      return Error("JobStandingOrderBase::setStandingOrder",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "day needed");
    }

    supported=false;
    for (k=0; k<dstr.length(); k++) {
      int m;

      m=atoi(dstr.substr(k, 1).c_str());
      if (m==0 || m==i) {
        supported=true;
        break;
      }
    } // for
    if (!supported) {
      DBG_ERROR(0, "day not supported by your bank");
      return Error("JobStandingOrderBase::setStandingOrder",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "day not supported by your bank");
    }
  }
  else if (toupper(*p)=='M') {
    tstr=GWEN_DB_GetCharValue(paramData(), "AllowedTurnusMonths", 0, "00");
    dstr=GWEN_DB_GetCharValue(paramData(), "AllowedMonthDays", 0, "00");

    // check day
    if (dstr.length()&1) {
      DBG_ERROR(0, "internal error: "
                "bad value length for AllowedMonthDays");
      return Error("JobStandingOrderBase::setStandingOrder",
                   ERROR_LEVEL_INTERNAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "bad value length for AllowedMonthDays");
    }
    i=GWEN_DB_GetIntValue(xa, "day", 0, -1);
    if (i==-1) {
      DBG_ERROR(0, "day needed");
      return Error("JobStandingOrderBase::setStandingOrder",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "day needed");
    }
    supported=false;
    for (k=0; k<dstr.length(); k+=2) {
      int m;

      m=atoi(dstr.substr(k, 2).c_str());
      DBG_ERROR(0, "Checking Day %d against %d", i, m);
      if (m==0 || m==i) {
        supported=true;
        break;
      }
    } // for
    if (!supported) {
      DBG_ERROR(0, "day not supported by your bank");
      return Error("JobStandingOrderBase::setStandingOrder",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "day not supported by your bank");
    }
  }
  else {
    DBG_ERROR(0, "unit must be either W or M");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "unit must be either W or M");
  }
  if (tstr.length()&1) {
    DBG_ERROR(0, "internal error: "
              "bad value length for AllowedTurnusMonths/Weeks");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_INTERNAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "bad value length for AllowedTurnusMonths/Weeks");
  }
  i=GWEN_DB_GetIntValue(xa, "turnus", 0, -1);
  if (i==-1) {
    DBG_ERROR(0, "turnus needed");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "turnus needed");
  }

  supported=false;
  for (k=0; k<tstr.length(); k+=2) {
    int m;

    m=atoi(tstr.substr(k, 2).c_str());
    if (m==0 || m==i) {
      supported=true;
      break;
    }
  } // for
  if (!supported) {
    DBG_ERROR(0, "turnus not supported by your bank");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "turnus not supported by your bank");
  }

  // check day
  supported=false;
  if (dstr.length()&1) {
    DBG_ERROR(0, "internal error: "
              "bad value length for AllowedMonthDays");
    return Error("JobStandingOrderBase::setStandingOrder",
                 ERROR_LEVEL_INTERNAL,
                 HBCI_ERROR_CODE_INVALID,
                 ERROR_ADVISE_DONTKNOW,
                 "bad value length for AllowedMonthDays");
  }

  // check textkey
  params=paramData();
  if (params) {
    GWEN_DB_NODE *n;
    int tk;

    tk=GWEN_DB_GetIntValue(xa, "textkey", 0, -1);
    if (tk==-1)
      tk=atoi(GWEN_DB_GetCharValue(xa, "textkey", 0, "-1"));
    if (tk==-1) {
      GWEN_DB_SetCharValue(xa, GWEN_DB_FLAGS_DEFAULT,
                           "textkey", "52");
      tk=52;
    }

    // check textkey
    n=GWEN_DB_GetGroup(params, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "textKey");
    if (n) {
      int i;

      for (i=0;;i++) {
        int ck;

        ck=GWEN_DB_GetIntValue(n, "key", i, -1);
        if (ck==-1) {
          if (i) {
            return Error("JobStandingOrderBase::setStandingOrder",
                         ERROR_LEVEL_NORMAL,
                         HBCI_ERROR_CODE_INVALID,
                         ERROR_ADVISE_DONTKNOW,
                         "textkey not supported for this account",
                         String::num2string(tk));
          }
          break;
        }
        if (ck==tk)
          break;
      } // for
    } // if textKey given
  } // if params found

  // check max purpose lines
  j=GWEN_DB_GetIntValue(paramData(), "MaxPurposeLines", 0, 0);
  if (j) {
    for (i=0; ; i++) {
      p=GWEN_DB_GetCharValue(xa, "Purpose", i, 0);
      if (!p)
        break;
    } // for
    if (i>j) {
      DBG_ERROR(0, "too many purpose lines");
      return Error("JobStandingOrderBase::setStandingOrder",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_INVALID,
                   ERROR_ADVISE_DONTKNOW,
                   "too many purpose lines");
    }
  }

  // TODO: check mindelay and maxdelay

  // standing order is ok, add it
  _sto=GWEN_DB_Group_dup(xa);
  addRequestData("", _sto);
  return Error();
}







/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobStartStandingOrder
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */



JobStartStandingOrder::JobStartStandingOrder(Pointer<Customer> c,
                                             Pointer<Account> acc,
                                             GWEN_DB_NODE *xa)
:JobStandingOrderBase("JobStartStandingOrder", c, acc, xa){
}



JobStartStandingOrder::JobStartStandingOrder(Pointer<Customer> c,
                                             Pointer<Account> acc)
:JobStandingOrderBase("JobStartStandingOrder", c, acc){
}



JobStartStandingOrder::~JobStartStandingOrder(){
}



Error JobStartStandingOrder::commit(){
  return JobStandingOrderBase::commit();
}



Error JobStartStandingOrder::evaluate(Pointer<Outbox> outbox){
  return JobStandingOrderBase::evaluate(outbox);
}






/* __________________________________________________________________________
 * AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
 *                      JobDeleteStandingOrder
 * YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
 */



JobDeleteStandingOrder::JobDeleteStandingOrder(Pointer<Customer> c,
                                               Pointer<Account> acc,
                                               GWEN_DB_NODE *xa)
:JobStandingOrderBase("JobDeleteStandingOrder", c, acc, xa){
}



JobDeleteStandingOrder::JobDeleteStandingOrder(Pointer<Customer> c,
                                               Pointer<Account> acc)
:JobStandingOrderBase("JobDeleteStandingOrder", c, acc){
}



JobDeleteStandingOrder::~JobDeleteStandingOrder(){
}



Error JobDeleteStandingOrder::commit(){
  return JobStandingOrderBase::commit();
}



Error JobDeleteStandingOrder::evaluate(Pointer<Outbox> outbox){
  return JobStandingOrderBase::evaluate(outbox);
}








