/***************************************************************************
  $RCSfile: accountjobs.h,v $
  -------------------
  cvs         : $Id: accountjobs.h,v 1.12 2004/04/02 05:28:29 aquamaniac Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifndef AQMONEY_ACCOUNTJOBS_H
#define AQMONEY_ACCOUNTJOBS_H

#include <aqmoney2/account.h>
#include <openhbci2.h>


/**
 * This is an abstract base class for account manipulating jobs.
 */
class AccountJob: public OutboxJob {
private:
  Pointer<Account> _account;
public:
  AccountJob(const string &name,
             Pointer<Customer> c,
             Pointer<Account> acc);
  virtual ~AccountJob();

  Pointer<Account> account() const { return _account;};
  const string &serverBankCode();

  AqMoneyAPI *api();

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   */
  virtual Error commit();

  /**
   *  Let the job check the result and set the flags.
   *
   * Let the job check the result. This means checking the results
   * of all Jobs involved (which otherwise is not done
   * automatically). This method sets its status and result
   * accordingly -- otherwise the status and result flags might not
   * be up-to-date.
   *
   * @author Martin Preuss<martin@libchipcard.de>
   * @return error object
   */
  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * @param sh if true the a short report will be created, otherwise
   * a full report is requested.
   */
  string resultString(bool sh);

  /**
   * Stores the result of a transaction within the given transaction.
   * It creates a subgroup named "jobresult" with the variables
   * "resultCode" and "resultText".
   */
  static void storeResult(GWEN_DB_NODE *rn, int code, const string &text);

};




/**
 *
 */
class JobGetBalance: public AccountJob {
private:

public:
  JobGetBalance(Pointer<Customer> c,
                Pointer<Account> acc);
  virtual ~JobGetBalance();

  virtual Error commit();

};



/**
 *
 */
class JobGetTransactions: public AccountJob {
private:

  Error _parseMT(const string &mt, GWEN_DB_NODE *rsp, const string &grpname);

public:
  JobGetTransactions(Pointer<Customer> c,
                     Pointer<Account> acc,
                     const Date &fromDate,
                     const Date &toDate);
  virtual ~JobGetTransactions();

  virtual Error commit();

};



/**
 *
 */
class JobGetStandingOrders: public AccountJob {
private:

public:
  JobGetStandingOrders(Pointer<Customer> c,
                       Pointer<Account> acc);
  virtual ~JobGetStandingOrders();

  virtual Error commit();

};




/**
 *
 */
class JobSingleTransferBase: public AccountJob {
private:
  GWEN_DB_NODE *_transaction;

public:
  /**
   * This constructor does NOT take over ownership of the given db.
   */
  JobSingleTransferBase(const string &name,
                        Pointer<Customer> c,
                        Pointer<Account> acc,
                        GWEN_DB_NODE *transaction);
  JobSingleTransferBase(const string &name,
                        Pointer<Customer> c,
                        Pointer<Account> acc);

  virtual ~JobSingleTransferBase();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * Returns a pointer to the transaction of this job.
   * You MUST NOT free this pointer !
   */
  GWEN_DB_NODE *transaction() { return _transaction;};

  /**
   * This method can be used in conjunction with the constructor which does
   * not take a transaction as a parameter.
   * You MUST not set a transaction more than once !
   * This method does NOT take over ownership of the given db.
   */
  Error setTransaction(GWEN_DB_NODE *transaction);
};



/**
 *
 */
class JobSingleTransfer: public JobSingleTransferBase {
private:
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobSingleTransfer(Pointer<Customer> c,
                    Pointer<Account> acc,
                    GWEN_DB_NODE *transaction);
  JobSingleTransfer(Pointer<Customer> c,
                    Pointer<Account> acc);
  virtual ~JobSingleTransfer();

  virtual Error commit();

};



/**
 *
 */
class JobSingleDebitNote: public JobSingleTransferBase {
private:
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobSingleDebitNote(Pointer<Customer> c,
                    Pointer<Account> acc,
                    GWEN_DB_NODE *transaction);
  JobSingleDebitNote(Pointer<Customer> c,
                     Pointer<Account> acc);
  virtual ~JobSingleDebitNote();

  virtual Error commit();

};



/**
 *
 */
class JobDatedSingleTransfer: public JobSingleTransferBase {
private:
  string _jobId;
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobDatedSingleTransfer(Pointer<Customer> c,
                         Pointer<Account> acc,
                         GWEN_DB_NODE *transaction);
  JobDatedSingleTransfer(Pointer<Customer> c,
                         Pointer<Account> acc);
  virtual ~JobDatedSingleTransfer();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * This id is assigned to the transfer by tha bank. You can use this id
   * to modify the transfer later (but before the execution date is due).
   */
  const string &jobId() const { return _jobId;};

};



/**
 *
 */
class JobChangeDatedSingleTransfer: public AccountJob {
private:
  GWEN_DB_NODE *_transfer;
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobChangeDatedSingleTransfer(Pointer<Customer> c,
                               Pointer<Account> acc,
                               GWEN_DB_NODE *transaction);
  JobChangeDatedSingleTransfer(Pointer<Customer> c,
                               Pointer<Account> acc);
  virtual ~JobChangeDatedSingleTransfer();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

  Error setTransfer(GWEN_DB_NODE *xa);

  GWEN_DB_NODE *transfer() { return _transfer;};

};



/**
 *
 */
class JobDeleteDatedSingleTransfer: public AccountJob {
private:
  GWEN_DB_NODE *_transfer;
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobDeleteDatedSingleTransfer(Pointer<Customer> c,
                               Pointer<Account> acc,
                               GWEN_DB_NODE *transaction);
  JobDeleteDatedSingleTransfer(Pointer<Customer> c,
                               Pointer<Account> acc);
  virtual ~JobDeleteDatedSingleTransfer();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

  Error setTransfer(GWEN_DB_NODE *xa);

  GWEN_DB_NODE *transfer() { return _transfer;};

};



/**
 *
 */
class JobGetDatedTransfers: public AccountJob {
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobGetDatedTransfers(Pointer<Customer> c,
                       Pointer<Account> acc,
                       const Date &fromDate,
                       const Date &toDate);
  virtual ~JobGetDatedTransfers();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);
};




/**
 *
 */
class JobMultiTransferBase: public AccountJob {
private:
  GWEN_DB_NODE *_transactions;
  string _defTextKey;
  unsigned int _transactionCount;
public:
  JobMultiTransferBase(const string &name,
                       const string &defTextKey,
                       Pointer<Customer> c,
                       Pointer<Account> acc);

  virtual ~JobMultiTransferBase();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);
  virtual Error addTransaction(GWEN_DB_NODE *xa);
  unsigned int transactionCount() const { return _transactionCount;};
  /**
   * Returns a pointer to the transaction of this job.
   * You MUST NOT free this pointer !
   */
  GWEN_DB_NODE *transactions() { return _transactions;};

};



/**
 *
 */
class JobMultiTransfer: public JobMultiTransferBase {
private:
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobMultiTransfer(Pointer<Customer> c,
                   Pointer<Account> acc);
  virtual ~JobMultiTransfer();

  virtual Error commit();

};



/**
 *
 */
class JobMultiDebitNote: public JobMultiTransferBase {
private:
public:
  /**
   * @param transaction transaction to be performed. The job stores and
   * uses a private copy of this transaction, so it does not take over
   * ownerschip of this object nor does it modify it.
   */
  JobMultiDebitNote(Pointer<Customer> c,
		    Pointer<Account> acc);
  virtual ~JobMultiDebitNote();

  virtual Error commit();

};



/**
 *
 */
class JobStandingOrderBase: public AccountJob {
private:
  GWEN_DB_NODE *_sto;

public:
  /**
   * This constructor does NOT take over ownership of the given db.
   */
  JobStandingOrderBase(const string &name,
                       Pointer<Customer> c,
                       Pointer<Account> acc,
                       GWEN_DB_NODE *transaction);
  JobStandingOrderBase(const string &name,
                       Pointer<Customer> c,
                       Pointer<Account> acc);

  virtual ~JobStandingOrderBase();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * Returns a pointer to the transaction of this job.
   * You MUST NOT free this pointer !
   */
  GWEN_DB_NODE *standingOrder() { return _sto;};

  /**
   * This method can be used in conjunction with the constructor which does
   * not take a transaction as a parameter.
   * You MUST not set a transaction more than once !
   * This method does NOT take over ownership of the given db.
   */
  virtual Error setStandingOrder(GWEN_DB_NODE *sto);
};



/**
 *
 */
class JobStartStandingOrder: public JobStandingOrderBase {
public:
  /**
   * This constructor does NOT take over ownership of the given db.
   */
  JobStartStandingOrder(Pointer<Customer> c,
                        Pointer<Account> acc,
                        GWEN_DB_NODE *transaction);
  JobStartStandingOrder(Pointer<Customer> c,
                        Pointer<Account> acc);

  virtual ~JobStartStandingOrder();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

};



/**
 *
 */
class JobDeleteStandingOrder: public JobStandingOrderBase {
public:
  /**
   * This constructor does NOT take over ownership of the given db.
   */
  JobDeleteStandingOrder(Pointer<Customer> c,
                        Pointer<Account> acc,
                        GWEN_DB_NODE *transaction);
  JobDeleteStandingOrder(Pointer<Customer> c,
                        Pointer<Account> acc);

  virtual ~JobDeleteStandingOrder();

  virtual Error commit();

  virtual Error evaluate(Pointer<Outbox> outbox);

};



#endif

