/***************************************************************************
  $RCSfile: adminjobs.h,v $
  -------------------
  cvs         : $Id: adminjobs.h,v 1.4 2004/01/25 18:51:22 aquamaniac Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifndef AQMONEY_ADMINJOBS_H
#define AQMONEY_ADMINJOBS_H


#include <openhbci2.h>
using namespace HBCI;


/**
 * This is an abstract base class for account manipulating jobs.
 * @author Martin Preuss<martin@libchipcard.de>
 */
class AdminJob: public OutboxJob {
public:
  AdminJob(const string &name,
           Pointer<Customer> c);
  virtual ~AdminJob();

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   */
  virtual Error commit();

  /**
   *  Let the job check the result and set the flags.
   *
   * Let the job check the result. This means checking the results
   * of all Jobs involved (which otherwise is not done
   * automatically). This method sets its status and result
   * accordingly -- otherwise the status and result flags might not
   * be up-to-date.
   *
   * @author Martin Preuss<martin@libchipcard.de>
   * @return error object
   */
  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * @param sh if true the a short report will be created, otherwise
   * a full report is requested.
   */
  string resultString(bool sh);

};



/**
 * Retrieves information about a bank anonymously (update BPD).
 * In most cases you won't need this job, since this information is updated
 * automatically by OpenHBCI on every contact with the bank server.
 * However, if you change the HBCI protocol version you should issue this job,
 * since some banks return different bank information depending on the HBCI
 * protocol version used.
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobGetBankInfo: public AdminJob {
public:
  JobGetBankInfo(Pointer<Customer> c);
  virtual ~JobGetBankInfo();
  virtual Error commit();
};



/**
 * Updates the information OpenHBCI has about this bank (update BPD).
 * You most probably need this after updating your XML files (because
 * then an update might deliver new BPD job descriptions)
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobUpdateBankInfo: public AdminJob {
public:
  JobUpdateBankInfo(Pointer<Customer> c);
  virtual ~JobUpdateBankInfo();
  virtual Error commit();
};




/**
 * This job retrieves the list of accounts the given customer is allowed to
 * manage via HBCI (update UPD).
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobGetAccounts: public AdminJob {
public:
  JobGetAccounts(Pointer<Customer> c);
  virtual ~JobGetAccounts();

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   * Adds the accounts to AqMoneys internal list of manageable accounts.
   */
  virtual Error commit();
};



typedef enum {
  AqmSyncModeSystemId=0,
  AqmSyncModeMsgNum,
  AqmSyncModeSignNum
} AQM_SYNC_MODE;

/**
 * @internal
 * This is an abstract job for retrieving the system id, last used message
 * number, last used signature number etc.
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobSync: public AdminJob {
private:
  int _syncmode;
public:
  JobSync(Pointer<Customer> c, AQM_SYNC_MODE syncmode);
  virtual ~JobSync();
};



/**
 * Retrieves the system id. This is needed in RDH mode (aka keyfile mode).
 * The system id is assigned to every user and every application (in theory)
 * and must be used on every further contact with the bank server.
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobGetSystemId: public JobSync {
private:
public:
  JobGetSystemId(Pointer<Customer> c);
  virtual ~JobGetSystemId();

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   */
  virtual Error commit();
};



/**
 * Retrieves the servers public keys.
 * This is needed in RDH mode (aka keyfile mode).
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobGetKeys: public AdminJob {
private:
  Pointer<RSAKey> _signKey;
  Pointer<RSAKey> _cryptKey;

public:
  JobGetKeys(Pointer<Customer> c);
  virtual ~JobGetKeys();

  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   */
  virtual Error commit();

  Pointer<RSAKey> signKey() { return _signKey;};
  Pointer<RSAKey> cryptKey() { return _cryptKey;};

};



/**
 * Sends the users public keys to the server.
 * This is needed in RDH mode (aka keyfile mode).
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobSendKeys: public AdminJob {
private:
  Pointer<RSAKey> _signKey;
  Pointer<RSAKey> _cryptKey;

public:
  JobSendKeys(Pointer<Customer> c,
              Pointer<RSAKey> signKey,
              Pointer<RSAKey> cryptKey);
  virtual ~JobSendKeys();

  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   */
  virtual Error commit();

};


/**
 * Sends the users public keys to the server.
 * This is needed in RDH mode (aka keyfile mode).
 * @author Martin Preuss<martin@libchipcard.de>
 */
class JobGetReports: public AdminJob {
private:
  Pointer<RSAKey> _signKey;
  Pointer<RSAKey> _cryptKey;

public:
  JobGetReports(Pointer<Customer> c);
  virtual ~JobGetReports();

  virtual Error evaluate(Pointer<Outbox> outbox);

  /**
   * This commits the result of this job's execution. This method is called
   * by the application after executing an outbox.
   */
  virtual Error commit();

};






#endif


