/***************************************************************************
  $RCSfile: aqmapi.cpp,v $
  -------------------
  cvs         : $Id: aqmapi.cpp,v 1.12 2004/02/17 21:37:55 rbla Exp $
  begin       : Tue Oct 14 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


/* Internationalization */
#ifdef ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define I18N(m) dgettext("aqmoney2",m)
#else
# define I18N(m) m
#endif

#include <gwenhywfar/directory.h>
#include <gwenhywfar/text.h>
#include <gwenhywfar/debug.h>
#include "aqmapi.h"
#include "storagefile.h"

#include "converter/converterdb.h"
#include "converter/converterold.h"
#include "converter/converterbin.h"
#include "converter/convertercsv.h"

#include <string.h>
#include <unistd.h>


AqMoneyAPI::AqMoneyAPI(bool readonly)
:API(readonly)
,_bankMessages(0){
  char homebuffer[256];
  string defaultPath;

  _bankMessages=GWEN_DB_Group_new("BankMessages");

  // create default path
  if (GWEN_Directory_GetHomeDirectory(homebuffer, sizeof(homebuffer))) {
    DBG_ERROR(0, "Buffer for home path too small");
    throw Error("AqMoneyAPI::loadEnvironment",
                ERROR_LEVEL_INTERNAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Buffer for home path too small");
  }
  defaultPath=homebuffer;
  defaultPath+="/.libaqmoney";
  setAapplicationDataDir(defaultPath);
  _bankManager=new BankManager(KTOBLZCHECK_DATAFILE);
}



AqMoneyAPI::~AqMoneyAPI(){
  GWEN_DB_Group_free(_bankMessages);
}




Error AqMoneyAPI::loadEnvironment(const string &openhbciFile,
                                  const string &aqmoneyFile){
  Error err;
  GWEN_DB_NODE *db;
  GWEN_DB_NODE *n;
  GWEN_DB_NODE *gr;

  DBG_INFO(0, "Reading environment");
  DBG_DEBUG(0, "  OpenHBCI file: \"%s\"", openhbciFile.c_str());
  DBG_DEBUG(0, "  AqMoney file: \"%s\"", aqmoneyFile.c_str());

  _defaultStorageType=STORAGEFILE_TYPENAME;
  _defaultStorageAddr="transactions";

  if (access(openhbciFile.c_str(), F_OK)) {
    DBG_WARN(0,
             "OpenHBCI file \"%s\" does not exist, will create it later",
             openhbciFile.c_str());
  }
  else {
    DBG_INFO(0, "Loading OpenHBCI file \"%s\"",
             openhbciFile.c_str());
    err=API::loadEnvironment(openhbciFile);
    if (!err.isOk())
      return Error("AqMoneyAPI::loadEnvironment", err);
  }

  if (access(aqmoneyFile.c_str(), F_OK)) {
    DBG_WARN(0,
             "AqMoney file \"%s\" does not exist, will create it later",
             aqmoneyFile.c_str());
  }
  else {
    DBG_INFO(0, "Loading AqMoney file \"%s\"",
             aqmoneyFile.c_str());
    db=GWEN_DB_Group_new("AqMoneyAPI::loadEnvironment");
    if (GWEN_DB_ReadFile(db, aqmoneyFile.c_str(),
			 GWEN_DB_FLAGS_DEFAULT|
			 GWEN_PATH_FLAGS_CREATE_GROUP)) {
      GWEN_DB_Group_free(db);
      return Error("AqMoneyAPI::loadEnvironment",
		   ERROR_LEVEL_NORMAL,
		   HBCI_ERROR_CODE_UNKNOWN,
		   ERROR_ADVISE_DONTKNOW,
		   "Error loading AqMoney config file",
		   aqmoneyFile);
    }

    try {
      _defaultStorageType=GWEN_DB_GetCharValue(db,
                                               "defaultStorage/type", 0,
                                               STORAGEFILE_TYPENAME);
      _defaultStorageAddr=GWEN_DB_GetCharValue(db,
                                               "defaultStorage/addr", 0,
                                               _defaultStorageAddr.c_str());

      // load accounts
      gr=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "accounts");
      if (gr) {
        n=GWEN_DB_GetFirstGroup(gr);
        while (n) {
          if (strcasecmp(GWEN_DB_GroupName(n), "account")==0) {
            // load account
            Pointer<Account> acc;

            acc=new Account(this, n);
            addAccount(acc);
          } // if account
          else {
            DBG_INFO(0, "Unkown section \"%s\", ignoring",
                     GWEN_DB_GroupName(n));
          }
          n=GWEN_DB_GetNextGroup(n);
        } // while n
      } // if "accounts"

      // load bank messages
      gr=GWEN_DB_GetGroup(db, GWEN_PATH_FLAGS_NAMEMUSTEXIST, "bankmsgs");
      if (gr) {
        n=GWEN_DB_GetFirstGroup(gr);
        while (n) {
          if (strcasecmp(GWEN_DB_GroupName(n), "bankMsg")==0) {
            GWEN_DB_AddGroup(_bankMessages, GWEN_DB_Group_dup(n));
          }
          else {
            DBG_INFO(0, "Unkown section \"%s\", ignoring",
                     GWEN_DB_GroupName(n));
          }
          n=GWEN_DB_GetNextGroup(n);
        } // while n
      } // if bank msgs

    } // try
    catch (Error xerr) {
        GWEN_DB_Group_free(db);
      return Error("AqMoneyAPI::loadEnvironment", xerr);
    }
    GWEN_DB_Group_free(db);
  }
  DBG_DEBUG(0, "Reading environment done.");
  return Error();
}



Error AqMoneyAPI::saveEnvironment(const string &openhbciFile,
                                  const string &aqmoneyFile){
  Error err;
  GWEN_DB_NODE *db;
  GWEN_DB_NODE *n;
  list<Pointer<Account> >::iterator ait;

  DBG_INFO(0, "Saving environment");
  DBG_DEBUG(0, "  OpenHBCI file: \"%s\"", openhbciFile.c_str());
  DBG_DEBUG(0, "  AqMoney file: \"%s\"", aqmoneyFile.c_str());

  err=API::saveEnvironment(openhbciFile);
  if (!err.isOk())
    return Error("AqMoneyAPI::saveEnvironment", err);

  db=GWEN_DB_Group_new("AqMoneyAPI::saveEnvironment");

  try {
    GWEN_DB_SetCharValue(db,
                         GWEN_DB_FLAGS_DEFAULT,
                         "defaultStorage/type",
                         _defaultStorageType.c_str());

    GWEN_DB_SetCharValue(db,
                         GWEN_DB_FLAGS_DEFAULT,
                         "defaultStorage/addr",
                         _defaultStorageAddr.c_str());

    DBG_DEBUG(0, "Saving accounts");
    for (ait=_accounts.begin(); ait!=_accounts.end(); ait++) {
      DBG_DEBUG(0, "Saving account %d/%s/%s",
                (*ait).ref().country(),
                (*ait).ref().bankCode().c_str(),
                (*ait).ref().accountId().c_str());
      n=GWEN_DB_GetGroup(db,
                         GWEN_DB_FLAGS_DEFAULT |
                         GWEN_PATH_FLAGS_CREATE_GROUP,
                         "accounts/account");
      assert(n);
      err=(*ait).ref().toConfig(n);
      if (!err.isOk()) {
        DBG_INFO(0, "Error saving account %d/%s/%s",
                 (*ait).ref().country(),
                 (*ait).ref().bankCode().c_str(),
                 (*ait).ref().accountId().c_str());
        GWEN_DB_Group_free(db);
        return Error("AqMoneyAPI::saveEnvironment", err);
      }
    } // for
    DBG_DEBUG(0, "Saving accounts done.");

    DBG_DEBUG(0, "Saving bank messages");
    n=GWEN_DB_GetFirstGroup(_bankMessages);
    while (n) {
      GWEN_DB_NODE *nn;

      nn=GWEN_DB_GetGroup(db,
                          GWEN_DB_FLAGS_DEFAULT |
                          GWEN_PATH_FLAGS_CREATE_GROUP,
                          "bankMsgs/bankMsg");
      assert(nn);
      GWEN_DB_AddGroupChildren(nn, n);
      n=GWEN_DB_GetNextGroup(n);
    } // while
    DBG_DEBUG(0, "Saving bank messages done.");

  } // try
  catch (Error xerr) {
    GWEN_DB_Group_free(db);
    return Error("AqMoneyAPI::loadEnvironment", xerr);
  }

  DBG_DEBUG(0, "Saving AqMoney file \"%s\"",
            aqmoneyFile.c_str());
  if (GWEN_DB_WriteFile(db, aqmoneyFile.c_str(),
                        GWEN_DB_FLAGS_DEFAULT)) {
    GWEN_DB_Group_free(db);
    return Error("AqMoneyAPI::saveEnvironment",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Error saving AqMoney config file",
                 aqmoneyFile);
  }

  GWEN_DB_Group_free(db);
  DBG_DEBUG(0, "Saving environment done.");
  return Error();

}



GWEN_DB_NODE *AqMoneyAPI::getBankMessages(){
  return GWEN_DB_Group_dup(_bankMessages);
}



Pointer<TransactionStorage>
AqMoneyAPI::storageFactory(const string &name,
                           int country,
                           const string &bankCode,
                           const string &account,
                           const string &addr,
                           const string &addrsuffix){
  Pointer<TransactionStorage> storage;

  if (strcasecmp(name.c_str(), STORAGEFILE_TYPENAME)==0) {
    // TransactionStorageFile
    storage=new TransactionStorageFile(this,
                                       addr,
                                       addrsuffix,
                                       country,
                                       bankCode,
                                       account);
  }
  else
    throw Error("AqMoneyAPI::storageFactory",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Unknown storage type",
                name);

  return storage;
}



Pointer<TransactionStorage>
AqMoneyAPI::storageFactory(GWEN_DB_NODE *stdb,
                           int country,
                           const string &bankCode,
                           const string &account){
  const char *p;
  Pointer<TransactionStorage> storage;

  try {
    p=GWEN_DB_GetCharValue(stdb, "type", 0, 0);
    if (!p)
      throw Error("AqMoneyAPI::storageFactory",
                  ERROR_LEVEL_NORMAL,
                  HBCI_ERROR_CODE_UNKNOWN,
                  ERROR_ADVISE_DONTKNOW,
                  "No storage type given");
    if (strcasecmp(p, STORAGEFILE_TYPENAME)==0) {
      // TransactionStorageFile
      storage=new TransactionStorageFile(this, stdb,
                                         country, bankCode, account);
    }
    else
      throw Error("AqMoneyAPI::storageFactory",
                  ERROR_LEVEL_NORMAL,
                  HBCI_ERROR_CODE_UNKNOWN,
                  ERROR_ADVISE_DONTKNOW,
                  "Unknown storage type",
                  p);
  }
  catch (Error xerr) {
    throw Error("AqMoneyAPI::storageFactory", xerr);
  }
  return storage;
}



void AqMoneyAPI::setDefaultStorage(const string &tname,
                                   const string &taddr){
  _defaultStorageType=tname;
  _defaultStorageAddr=taddr;
}



Error AqMoneyAPI::processOutboxResponse(GWEN_DB_NODE *db){
  GWEN_DB_NODE *n;
  Error err;
  string custId;

  assert(db);
  n=GWEN_DB_GetFirstGroup(db);
  while (n){
    if (strcasecmp(GWEN_DB_GroupName(n), "accountdata")==0) {
      // found account data, create account
      string accountId;
      string accountSubId;
      string bankCode;
      int country;
      Pointer<Account> acc;
      string custid;
      list<Pointer<Customer> > customers;
      Pointer<User> user;
      Pointer<Bank> bank;

      country=GWEN_DB_GetIntValue(n, "country", 0, 280);
      custid=GWEN_DB_GetCharValue(n, "customer", 0, "");
      bankCode=GWEN_DB_GetCharValue(n, "bankcode", 0, "");
      accountId=GWEN_DB_GetCharValue(n, "accountid", 0, "");
      accountSubId=GWEN_DB_GetCharValue(n, "accountsubid", 0, "");
      if (bankCode.empty() || accountId.empty() || custid.empty()) {
        DBG_ERROR(0, "AccountData without bank code/account id/customer id");
        GWEN_DB_Group_free(db);
        return Error("AqMoneyAPI::processOutboxResponse",
                     ERROR_LEVEL_NORMAL,
                     HBCI_ERROR_CODE_UNKNOWN,
                     ERROR_ADVISE_DONTKNOW,
                     "Some vars are missing");
      }

      bank=findBank(country, bankCode);
      if (!bank.isValid()) {
        /* bank uses a different bank code for the accounts, so find
         * the matching bank for the given customer */
        customers=getCustomers(country, "*", custid);
        if (customers.empty()) {
          DBG_ERROR(0, "Unknown customer %d/%s/%s",
                    country, bankCode.c_str(), custid.c_str());
          GWEN_DB_Group_free(db);
          return Error("AqMoneyAPI::processOutboxResponse",
                       ERROR_LEVEL_NORMAL,
                       HBCI_ERROR_CODE_UNKNOWN,
                       ERROR_ADVISE_DONTKNOW,
                       "Unknown customer");
        }
        if (customers.size()!=1) {
          DBG_ERROR(0, "Ambiguous customer %d/%s/%s",
                    country, bankCode.c_str(), custid.c_str());
          GWEN_DB_Group_free(db);
          return Error("AqMoneyAPI::processOutboxResponse",
                       ERROR_LEVEL_NORMAL,
                       HBCI_ERROR_CODE_UNKNOWN,
                       ERROR_ADVISE_DONTKNOW,
                       "Ambiguous customer");
        }
        user=customers.front().ref().user();
        bank=user.ref().bank();
      }
      acc=findAccount(country, bankCode, accountId, accountSubId);
      if (acc.isValid()) {
        const char *p;

        // Update account information
        DBG_INFO(0, "Account %d/%s/%s already exists, updating",
                 country, bankCode.c_str(), accountId.c_str());

        p=GWEN_DB_GetCharValue(n, "name", 0, 0);
        if (p)
          acc.ref().setAccountName(p);
        p=GWEN_DB_GetCharValue(n, "customer", 0, 0);
        if (p)
          acc.ref().setCustomer(p);
        p=GWEN_DB_GetCharValue(n, "currency", 0, 0);
        if (p)
          acc.ref().setCurrency(p);
        p=GWEN_DB_GetCharValue(n, "name1", 0, 0);
        if (p)
          acc.ref().setUserName(p);
      }
      else {
        // new account, create and add it
        try {
          DBG_INFO(0, "Creating account %d/%s/%s",
                   country, bankCode.c_str(), accountId.c_str());
          acc=new Account(this,
			  accountId,
			  accountSubId,
                          country,
                          bankCode,
                          bank.ref().bankCode(),
                          GWEN_DB_GetCharValue(n, "name", 0, ""),
                          GWEN_DB_GetCharValue(n, "customer", 0, ""),
                          GWEN_DB_GetCharValue(n, "currency", 0, "EUR"),
                          GWEN_DB_GetCharValue(n, "name1", 0, ""),
                          _defaultStorageType,
                          _defaultStorageAddr);
        }
        catch (Error xerr) {
          DBG_INFO(0, "Could not create account");
          GWEN_DB_Group_free(db);
          return Error("AqMoneyAPI::processOutboxResponse", xerr);
        }
        err=addAccount(acc);
        DBG_INFO(0, "Added account %d/%s/%s",
                 country, bankCode.c_str(), accountId.c_str());
      } // if account doesnt exist
    } // if "AccountData"
    else if (strcasecmp(GWEN_DB_GroupName(n), "bankmsg")==0) {
      // add to existing bank messages
      GWEN_DB_AddGroup(_bankMessages, GWEN_DB_Group_dup(n));
      fprintf(stderr, I18N("------------------------------------\n"
                           "Message from \"%s\":\n"
                           "Subject: %s\n"
                           "\n%s\n\n"),
                      GWEN_DB_GetCharValue(n, "bankCode", 0, I18N("<unknown>")),
                      GWEN_DB_GetCharValue(n, "subject", 0, I18N("<empty>")),
                      GWEN_DB_GetCharValue(n, "text", 0, I18N("<empty>"))
             );
    }
    n=GWEN_DB_GetNextGroup(n);
  } // while n

  GWEN_DB_Group_free(db);
  return Error();
}



Error AqMoneyAPI::addAccount(Pointer<Account> acc){
  if (findAccount(acc.ref().country(),
                  acc.ref().bankCode(),
		  acc.ref().accountId(),
		  acc.ref().accountSubId()).isValid()) {
    return Error("AqMoneyAPI::addAccount",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Account already exists");
  }
  _accounts.push_back(acc);
  return Error();
}



Pointer<Account> AqMoneyAPI::findAccount(int country,
					 const string &bankCode,
					 const string &accountId,
					 const string &accountSubId){
  list<Pointer<Account> >::iterator ait;

  for (ait=_accounts.begin(); ait!=_accounts.end(); ait++) {
    if ((*ait).ref().country()==country)
      if (strcasecmp((*ait).ref().bankCode().c_str(), bankCode.c_str())==0)
	if (strcasecmp((*ait).ref().accountId().c_str(),
		       accountId.c_str())==0)
	  if (strcasecmp((*ait).ref().accountSubId().c_str(),
			 accountSubId.c_str())==0){
	    DBG_INFO(0, "Account %d/%s/%s found",
		     country, bankCode.c_str(), accountId.c_str());
	    return *ait;
	  }
  } // for
  DBG_INFO(0, "Account %d/%s/%s not found",
	   country, bankCode.c_str(), accountId.c_str());
  return 0;
}



list<Pointer<Account> > AqMoneyAPI::getAccounts(int country,
                                                const string &bankCode,
						const string &accountId,
						const string &accountSubId){
  list<Pointer<Account> > accs;
  list<Pointer<Account> >::iterator ait;

  for (ait=_accounts.begin(); ait!=_accounts.end(); ait++) {
    if ((*ait).ref().country()==country || country==0)
      if (-1!=GWEN_Text_ComparePattern((*ait).ref().bankCode().c_str(),
                                       bankCode.c_str(), false))
        if (-1!=GWEN_Text_ComparePattern((*ait).ref().accountId().c_str(),
					 accountId.c_str(), false))
	  if (-1!=GWEN_Text_ComparePattern((*ait).ref().accountSubId().c_str(),
					   accountSubId.c_str(), false)) {
	    DBG_INFO(0, "Account %d/%s/%s matches",
		     (*ait).ref().country(),
		     (*ait).ref().bankCode().c_str(),
		     (*ait).ref().accountId().c_str());
	    accs.push_back(*ait);
	  }
  } // for

  return accs;
}



Pointer<AQMConverter> AqMoneyAPI::converterFactory(const string &name,
                                                   const string &param){
  Pointer<AQMConverter> c;

  if (strcasecmp(name.c_str(), "db")==0)
    c=new AQMConverterDB();
  else if (strcasecmp(name.c_str(), "old")==0)
    c=new AQMConverterOld(param);
  else if (strcasecmp(name.c_str(), "dtaus")==0)
    c=new AQMConverterBin(this, name, param);
  else if (strcasecmp(name.c_str(), "swift")==0)
    c=new AQMConverterBin(this, name, param);
  else if (strcasecmp(name.c_str(), "csv")==0)
    c=(param.length() != 0)?new AQMConverterCSV(param):new AQMConverterCSV();
  else {
    DBG_ERROR(0, "Unknown converter type \"%s\"", name.c_str());
  }

  return c;
}



Pointer<Account> AqMoneyAPI::accountFactory(const string &customer,
                                            const string &userName,
					    const string &accountId,
					    const string &accountSubId,
                                            int country,
                                            const string &bankCode,
                                            const string &serverBankCode,
                                            const string &accountName,
                                            const string &currency,
                                            const string &storageType,
                                            const string &storageAddr){
  Pointer<Account> acc;

  acc=new Account(this,
                  accountId,
		  accountSubId,
                  country,
                  bankCode,
                  ((serverBankCode.empty())?bankCode:serverBankCode),
                  accountName,
                  customer,
                  currency,
                  userName,
                  ((storageType.empty())?_defaultStorageType:storageType),
                  ((storageAddr.empty())?_defaultStorageAddr:storageAddr));
  return acc;
}







