/***************************************************************************
 $RCSfile: oldconf.c,v $
                             -------------------
    cvs         : $Id: oldconf.c,v 1.3 2004/01/25 18:51:29 aquamaniac Exp $
    begin       : Mon Dec 02 2002
    copyright   : (C) 2002 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "oldconf.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>


#ifdef MEMTRACE
static int CONFIGOLDVARIABLE_Counter=0;
static int CONFIGOLDGROUP_Counter=0;
#endif


CONFIGOLDVALUE *ConfigOld__Value_new(const char *value){
  CONFIGOLDVALUE *v;

  v=(CONFIGOLDVALUE*)malloc(sizeof(CONFIGOLDVALUE));
  assert(v);
  memset(v, 0, sizeof(CONFIGOLDVALUE));
  if (value) {
    v->value=(char*)malloc(strlen(value)+1);
    assert(v->value);
    memmove(v->value, value, strlen(value)+1);
  }
  return v;
}


void ConfigOld__Value_free(CONFIGOLDVALUE *v){
  if (v) {
    if (v->value)
      free(v->value);
    free(v);
  }
}


CONFIGOLDVARIABLE *ConfigOld__Variable_new(const char *name,
				     const char *value){
  CONFIGOLDVARIABLE *v;

  v=(CONFIGOLDVARIABLE*)malloc(sizeof(CONFIGOLDVARIABLE));
  assert(v);
  memset(v, 0, sizeof(CONFIGOLDVARIABLE));
  if (name) {
    v->name=(char*)malloc(strlen(name)+1);
    assert(v->name);
    memmove(v->name, name, strlen(name)+1);
  }
  if (value)
    v->values=ConfigOld__Value_new(value);
#ifdef MEMTRACE
  CONFIGOLDVARIABLE_Counter++;
  DBG_INFO(0, "Created variable, now %d",CONFIGOLDVARIABLE_Counter);
#endif
  return v;
}


void ConfigOld__Variable_Clear(CONFIGOLDVARIABLE *v) {
  assert(v);
  if (v->values) {
    CONFIGOLDVALUE *val;
    CONFIGOLDVALUE *next;

    val=v->values;
    while(val) {
      next=val->next;
      ConfigOld__Value_free(val);
      val=next;
    } /* while */
  } /* if values */
}


void ConfigOld__Variable_free(CONFIGOLDVARIABLE *v){
  DBG_ENTER;
  if (v) {
    ConfigOld__Variable_Clear(v);
    free(v->name);
    free(v);
#ifdef MEMTRACE
    CONFIGOLDVARIABLE_Counter--;
    DBG_INFO(0, "Freed variable, now %d",CONFIGOLDVARIABLE_Counter);
#endif
  }
  DBG_LEAVE;
}


CONFIGOLDGROUP * ConfigOld__Group_new(const char *name){
  CONFIGOLDGROUP *g;

  DBG_VERBOUS(0, "Creating group \"%s\"",name);
  g=(CONFIGOLDGROUP*)malloc(sizeof(CONFIGOLDGROUP));
  assert(g);
  memset(g, 0, sizeof(CONFIGOLDGROUP));
  if (name) {
    g->name=(char*)malloc(strlen(name)+1);
    assert(g->name);
    memmove(g->name, name, strlen(name)+1);
  }
#ifdef MEMTRACE
  CONFIGOLDGROUP_Counter++;
  DBG_NOTICE("Created group, now %d",CONFIGOLDGROUP_Counter);
#endif
  return g;
}


void ConfigOld__Group_free(CONFIGOLDGROUP *g);

void ConfigOld__Group_Clear(CONFIGOLDGROUP *g) {
  CONFIGOLDGROUP *currg, *nextg;
  CONFIGOLDVARIABLE *currv, *nextv;

  assert(g);

  /* free all variables */
  currv=g->variables;
  while(currv) {
    nextv=currv->next;
    ConfigOld__Variable_free(currv);
    currv=nextv;
  } /* while */
  g->variables=0;

  /* free all subgroups */
  currg=g->groups;
  while(currg) {
    nextg=currg->next;
    ConfigOld__Group_free(currg);
    currg=nextg;
  } /* while */
  g->groups=0;

}


void ConfigOld__Group_free(CONFIGOLDGROUP *g){
  DBG_ENTER;
  if (g) {
    ConfigOld__Group_Clear(g);
    free(g->name);
    free(g);
#ifdef MEMTRACE
    CONFIGOLDGROUP_Counter--;
    DBG_NOTICE("Freed group, now %d",CONFIGOLDGROUP_Counter);
#endif
  }
  DBG_LEAVE;
}


void ConfigOld_AddGroup(CONFIGOLDGROUP *p, CONFIGOLDGROUP *w){
  CONFIGOLDGROUP *curr;

  DBG_ENTER;
  assert(p);
  assert(w);

  w->parent=p;
  curr=p->groups;
  if (!curr) {
    p->groups=w;
  }
  else {
    /* find last */
    while(curr->next) {
      curr=curr->next;
    } /* while */
    curr->next=w;
  }
  DBG_LEAVE;
}


void ConfigOld_UnlinkGroup(CONFIGOLDGROUP *p, CONFIGOLDGROUP *w){
  CONFIGOLDGROUP *curr;

  DBG_ENTER;
  assert(p);
  assert(w);

  curr=p->groups;
  if (curr) {
    if (curr==w) {
      p->groups=curr->next;
    }
    else {
      /* find predecessor */
      while(curr->next!=w) {
	curr=curr->next;
      } /* while */
      if (curr)
	curr->next=w->next;
    }
  }
  w->parent=0;
  w->next=0;
  DBG_LEAVE;
}


CONFIGOLDGROUP *ConfigOld__FindGroup(CONFIGOLDGROUP *p, const char *name){
  CONFIGOLDGROUP *curr;

  DBG_ENTER;
  assert(p);
  assert(name);
  curr=p->groups;

  DBG_VERBOUS(0, "Find group \"%s\"",name);
  while(curr) {
    if (curr->name) {
      if (ConfigOld_Compare(curr->name, name)==0) {
	DBG_DEBUG(0, "Group \"%s\" found", name);
	DBG_LEAVE;
	return curr;
      }
    }
    curr=curr->next;
  } /* while */
  DBG_DEBUG(0, "Group \"%s\" not found", name);
  DBG_LEAVE;
  return 0;
}


void ConfigOld__AddVariable(CONFIGOLDGROUP *p, CONFIGOLDVARIABLE *v){
  CONFIGOLDVARIABLE *curr;

  DBG_ENTER;
  assert(p);
  assert(v);

  v->parent=p;
  curr=p->variables;
  if (!curr) {
    p->variables=v;
  }
  else {
    /* find last */
    while(curr->next) {
      curr=curr->next;
    } /* while */
    curr->next=v;
  }
  DBG_LEAVE;
}


void ConfigOld__UnlinkVariable(CONFIGOLDGROUP *p, CONFIGOLDVARIABLE *w){
  CONFIGOLDVARIABLE *curr;

  DBG_ENTER;
  assert(p);
  assert(w);

  curr=p->variables;
  if (curr) {
    if (curr==w) {
      p->variables=curr->next;
    }
    else {
      /* find predecessor */
      while(curr->next!=w) {
	curr=curr->next;
      } /* while */
      if (curr)
	curr->next=w->next;
    }
  }
  w->parent=0;
  w->next=0;
  DBG_LEAVE;
}


CONFIGOLDVARIABLE *ConfigOld__FindVariable(CONFIGOLDGROUP *p, const char *name){
  CONFIGOLDVARIABLE *curr;

  DBG_ENTER;
  assert(p);
  assert(name);
  curr=p->variables;

  while(curr) {
    if (curr->name)
      if (ConfigOld_Compare(curr->name, name)==0) {
	DBG_DEBUG(0, "Variable \"%s\" found", name);
        DBG_LEAVE;
	return curr;
      }
    curr=curr->next;
  } /* while */
  DBG_DEBUG(0, "Variable \"%s\" not found", name);
  DBG_LEAVE;
  return 0;
}


void ConfigOld__AddValue(CONFIGOLDVARIABLE *p, CONFIGOLDVALUE *v){
  CONFIGOLDVALUE *curr;

  DBG_ENTER;
  assert(p);
  assert(v);

  curr=p->values;
  if (!curr) {
    p->values=v;
  }
  else {
    /* find last */
    while(curr->next) {
      curr=curr->next;
    } /* while */
    curr->next=v;
  }
  DBG_LEAVE;
}


void ConfigOld__UnlinkValue(CONFIGOLDVARIABLE *p, CONFIGOLDVALUE *w){
  CONFIGOLDVALUE *curr;

  DBG_ENTER;
  assert(p);
  assert(w);

  curr=p->values;
  if (curr) {
    if (curr==w) {
      p->values=curr->next;
    }
    else {
      /* find predecessor */
      while(curr->next!=w) {
	curr=curr->next;
      } /* while */
      if (curr)
	curr->next=w->next;
    }
  }
  w->next=0;
  DBG_LEAVE;
}


void ConfigOld__SetValue(CONFIGOLDVARIABLE *p, const char *value){
  assert(p);

  ConfigOld__Variable_Clear(p);
  if (value)
    p->values=ConfigOld__Value_new(value);
}


CONFIGOLDVALUE *ConfigOld__FindValue(CONFIGOLDVARIABLE *p, const char *value){
  CONFIGOLDVALUE *curr;

  DBG_ENTER;
  assert(p);
  assert(value);
  curr=p->values;

  while(curr) {
    if (curr->value)
      if (ConfigOld_Compare(curr->value, value)==0) {
	DBG_DEBUG(0, "Value \"%s\" found", value);
	DBG_LEAVE;
	return curr;
      }
    curr=curr->next;
  } /* while */
  DBG_DEBUG(0, "Value \"%s\" not found", value);
  DBG_LEAVE;
  return 0;
}


void *ConfigOld__GetPath(CONFIGOLDGROUP *root,
		      const char *path,
		      unsigned int mode) {
  CONFIGOLDGROUP *curr;
  CONFIGOLDVARIABLE *var;
  char currname[256];
  const char *psrc;
  char *pdest;
  int i;

  assert(root);
  assert(path);

  psrc=path;
  DBG_VERBOUS(0, "Getting path \"%s\"",path);
  if (*psrc=='/')
    psrc++;

  while(*psrc) {
    /* get path element */
    pdest=currname;
    i=sizeof(currname)-1;
    currname[0]=0;
    while(*psrc && i--) {
      if (*psrc=='/')
	break;
      *pdest=*psrc;
      psrc++;
      pdest++;
    } /* while */
    if (!i) {
      DBG_ERROR(0, "Path element too long (limit is %d chars)",
                sizeof(currname))
	return 0;
    }
    *pdest=0;
    if (*psrc)
      psrc++;

    if (*psrc==0) {
      /* handling name (last path element) */
      if (mode & CONFIGOLDMODE_VARIABLE) {
	DBG_VERBOUS(0, "Want variable");
	/* want a variable */
	if ((mode & CONFIGOLDMODE_NAMECREATE_VARIABLE) ||
	    (mode & CONFIGOLDMODE_PATHCREATE)) {
	  DBG_VERBOUS(0, "Forced to create variable \"%s\" in \"%s\"",
		   currname, root->name);
	  var=ConfigOld__Variable_new(currname,0);
	  ConfigOld__AddVariable(root, var);
	}
	else {
	  var=ConfigOld__FindVariable(root, currname);
	  if (!var) {
	    if (mode & CONFIGOLDMODE_NAMEMUSTEXIST) {
	      DBG_DEBUG(0, "Path \"%s\" not found (%s)",path, currname);
	      return 0;
	    }
	    DBG_VERBOUS(0, "Variable \"%s\" does not exists, creating it",
		      currname);
	    var=ConfigOld__Variable_new(currname,0);
	    ConfigOld__AddVariable(root, var);
	  }
	  else {
	    if (mode & CONFIGOLDMODE_NAMEMUSTNOTEXIST) {
	      DBG_DEBUG(0, "Path \"%s\" already exists (%s)",path, currname);
	      return 0;
	    }
	  }
	}
	DBG_DEBUG(0, "Path \"%s\" available", path);
	DBG_LEAVE;
	return var;
      } /* if wantVariable */
      else {
	DBG_VERBOUS(0, "Want group");
        /* want a group */
	if ((mode & CONFIGOLDMODE_NAMECREATE_GROUP) ||
	    (mode & CONFIGOLDMODE_PATHCREATE)) {
	  DBG_DEBUG(0, "Forced to create group \"%s\" in \"%s\"",
		   currname, root->name);
	  curr=ConfigOld__Group_new(currname);
	  ConfigOld_AddGroup(root, curr);
	}
	else {
	  curr=ConfigOld__FindGroup(root, currname);
	  if (!curr) {
	    if (mode & CONFIGOLDMODE_NAMEMUSTEXIST) {
	      DBG_DEBUG(0, "Path \"%s\" not found (%s)",path, currname);
	      return 0;
	    }
	    DBG_DEBUG(0, "Group \"%s\" does not exists in \"%s\", creating it",
		     currname, root->name);
	    curr=ConfigOld__Group_new(currname);
	    ConfigOld_AddGroup(root, curr);
	  }
	  else {
	    if (mode & CONFIGOLDMODE_NAMEMUSTNOTEXIST) {
	      DBG_DEBUG(0, "Path \"%s\" already exists (%s)",path, currname);
	      return 0;
	    }
	  }
	}
	DBG_VERBOUS(0, "Path \"%s\" available", path);
	if (mode & CONFIGOLDMODE_OVERWRITE_GROUPS)
	  ConfigOld_ClearGroup(curr);
	DBG_LEAVE;
	return curr;
      } /* "else" of "if wantVariable" */
    } /* if *path==0 */
    else {
      /* not at the end of the path */
      if (mode & CONFIGOLDMODE_PATHCREATE) {
	DBG_VERBOUS(0, "Forced to create group \"%s\" in \"%s\"",
		 currname, root->name);
	curr=ConfigOld__Group_new(currname);
	ConfigOld_AddGroup(root, curr);
      }
      else {
	curr=ConfigOld__FindGroup(root, currname);
	if (!curr) {
	  if (mode & CONFIGOLDMODE_PATHMUSTEXIST) {
	    DBG_DEBUG(0, "Path \"%s\" not found (%s)",path, currname);
	    return 0;
	  }
	  DBG_VERBOUS(0, "Group \"%s\" does not exists, creating it",currname);
	  curr=ConfigOld__Group_new(currname);
	  ConfigOld_AddGroup(root, curr);
	}
	else {
	  if (mode & CONFIGOLDMODE_PATHMUSTNOTEXIST) {
	    DBG_DEBUG(0, "Path \"%s\" already exists (%s)",path, currname);
	    return 0;
	  }
	}
      }
      root=curr;
    }
  } /* while */
  DBG_ERROR(0, "Uuups :-} We should never reach this point (\"%s\")", path);
  return 0;
}


CONFIGOLDVALUE *ConfigOld__Value_duplicate(CONFIGOLDVALUE *p){
  CONFIGOLDVALUE *copy;

  assert(p);
  DBG_VERBOUS(0, "Duplicating value \"%s\"",p->value);
  copy=ConfigOld__Value_new(p->value);
  return copy;
}


CONFIGOLDVARIABLE *ConfigOld__Variable_duplicate(CONFIGOLDVARIABLE *p){
  CONFIGOLDVARIABLE *copy;
  CONFIGOLDVALUE *val1, *val2;

  assert(p);
  DBG_VERBOUS(0, "Duplicating variable \"%s\"",p->name);
  copy=ConfigOld__Variable_new(p->name,0);

  val1=p->values;
  while(val1) {
    val2=ConfigOld__Value_duplicate(val1);
    ConfigOld__AddValue(copy,val2);
    val1=val1->next;
  }
  return copy;
}


CONFIGOLDGROUP *ConfigOld__Group_duplicate(CONFIGOLDGROUP *p){
  CONFIGOLDGROUP *copy;
  CONFIGOLDVARIABLE *var1, *var2;
  CONFIGOLDGROUP *g1, *g2;

  assert(p);
  DBG_VERBOUS(0, "Duplicating group \"%s\"",p->name);
  copy=ConfigOld__Group_new(p->name);

  var1=p->variables;
  while(var1) {
    var2=ConfigOld__Variable_duplicate(var1);
    ConfigOld__AddVariable(copy,var2);
    var1=var1->next;
  }

  g1=p->groups;
  while(g1) {
    g2=ConfigOld__Group_duplicate(g1);
    ConfigOld_AddGroup(copy, g2);
    g1=g1->next;
  }
  return copy;
}


void ConfigOld_ImportTreeChildren(CONFIGOLDGROUP *p, CONFIGOLDGROUP *g) {
  CONFIGOLDGROUP *currg, *newg;
  CONFIGOLDVARIABLE *currv, *newv;

  assert(p);
  assert(g);

  /* add all variables */
  currv=g->variables;
  while(currv) {
    newv=ConfigOld__Variable_duplicate(currv);
    ConfigOld__AddVariable(p, newv);
    currv=currv->next;
  } /* while */

  /* free all subgroups */
  currg=g->groups;
  while(currg) {
    newg=ConfigOld__Group_duplicate(currg);
    ConfigOld_AddGroup(p, newg);
    currg=currg->next;
  } /* while */
}









const char *ConfigOld_GetValue(CONFIGOLDGROUP *root,
			    const char *path,
			    const char *defaultValue,
			    int idx){
  CONFIGOLDVARIABLE *var;
  CONFIGOLDVALUE *val;

  assert(root);
  assert(path);
  DBG_VERBOUS(0, "GetValue for \"%s\"",path);
  var=ConfigOld__GetPath(root,path,
		      CONFIGOLDMODE_PATHMUSTEXIST |
		      CONFIGOLDMODE_NAMEMUSTEXIST |
		      CONFIGOLDMODE_VARIABLE);
  if (!var) {
    DBG_DEBUG(0, "Returning default value for path \"%s\"",path);
    return defaultValue;
  }
  val=var->values;
  while (val && idx) {
    val=val->next;
    idx--;
  }
  if (!val) {
    DBG_DEBUG(0, "Returning default value for path \"%s\"",path);
    return defaultValue;
  }

  return val->value;;
}


int ConfigOld_GetIntValue(CONFIGOLDGROUP *root,
		       const char *path,
		       int defaultValue,
		       int idx){
  const char *p;
  int val;

  p=ConfigOld_GetValue(root, path, 0, idx);
  if (!p) {
    DBG_DEBUG(0, "Returning default value for \"%s\"", path);
    return defaultValue;
  }
  if (sscanf(p,"%d",&val)!=1) {
    DBG_INFO(0, "Bad value for \"%s\", will return default value instead",
	      path);
    return defaultValue;
  }
  return val;
}


int ConfigOld_GetBoolValue(CONFIGOLDGROUP *root,
			const char *path,
			int defaultValue,
			int idx){
  const char *p;
  int val;

  p=ConfigOld_GetValue(root, path, 0, idx);
  if (!p) {
    DBG_DEBUG(0, "Returning default value for \"%s\"", path);
    return defaultValue;
  }
  if (sscanf(p,"%i",&val)==1)
    return (val!=0);
  else if (strcasecmp(p, "yes")==0)
    return 1;
  else if (strcasecmp(p, "no")==0)
    return 0;
  else if (strcasecmp(p, "true")==0)
    return 1;
  else if (strcasecmp(p, "false")==0)
    return 0;
  else if (strcasecmp(p, "yeah")==0)
    return 1;
  else if (strcasecmp(p, "nope")==0)
    return 0;

  return defaultValue;
}



time_t ConfigOld_GetTimeValue(CONFIGOLDGROUP *root,
			   const char *path,
			   time_t defaultValue,
			   int idx){
  char buffer[128];
  struct tm tm;
  struct tm *tt;
  const char *p;
  time_t val;
  int year;

  p=ConfigOld_GetValue(root, path, 0, idx);
  if (!p) {
    DBG_INFO(0, "Returning default value for \"%s\"", path);
    return defaultValue;
  }

  val=time(0);
  tt=localtime(&val);
  memmove(&tm, tt, sizeof(tm));

  buffer[0]=0;
  if (sscanf(p,"%d/%d/%d-%d:%d:%d",
	     &year,
	     &tm.tm_mon,
	     &tm.tm_mday,
	     &tm.tm_hour,
	     &tm.tm_min,
	     &tm.tm_sec)!=6) {
    DBG_INFO(0, "Invalid time, returning default value for \"%s\"", path);
    return defaultValue;
  }
  tm.tm_year=year-1900;
  tm.tm_isdst=-1;
  val=mktime(&tm);
  return val;
}





int ConfigOld_SetValue(CONFIGOLDGROUP *root,
		    unsigned int mode,
		    const char *path,
		    const char *value){
  CONFIGOLDVARIABLE *var;

  assert(root);
  assert(path);

  DBG_VERBOUS(0, "SetValue for \"%s\"",path);
  var=(CONFIGOLDVARIABLE*)ConfigOld__GetPath(root,path,
				       mode | CONFIGOLDMODE_VARIABLE);
  if (!var) {
    DBG_DEBUG(0, "Path not available");
    return 1;
  }
  if (mode & CONFIGOLDMODE_OVERWRITE_VARS) {
    ConfigOld__Variable_Clear(var);
    if (value) {
      if (*value || (mode &CONFIGOLDMODE_EMPTY_VARIABLES))
	var->values=ConfigOld__Value_new(value);
    }
  }
  else {
    CONFIGOLDVALUE *val;
    if (value) {
      if (*value || (mode &CONFIGOLDMODE_EMPTY_VARIABLES)) {
	val=ConfigOld__Value_new(value);
	ConfigOld__AddValue(var, val);
      }
    }
  }
  return 0;
}


int ConfigOld_SetIntValue(CONFIGOLDGROUP *root,
		       unsigned int mode,
		       const char *path,
		       int value){
  char buffer[32];

  buffer[0]=0;
  sprintf(buffer,"%d",value);
  return ConfigOld_SetValue(root,
			 mode,
			 path,
			 buffer);
}


int ConfigOld_SetBoolValue(CONFIGOLDGROUP *root,
			unsigned int mode,
			const char *path,
			int value){
  return ConfigOld_SetValue(root,
			 mode,
			 path,
			 value?"yes":"no");
}


int ConfigOld_SetTimeValue(CONFIGOLDGROUP *root,
			unsigned int mode,
			const char *path,
			time_t value){
  char buffer[128];
  struct tm *tm;

  tm=localtime(&value);
  assert(tm);

  buffer[0]=0;
  sprintf(buffer,"%4d/%02d/%02d-%02d:%02d:%02d",
	  tm->tm_year+1900,
	  tm->tm_mon,
	  tm->tm_mday,
	  tm->tm_hour,
	  tm->tm_min,
	  tm->tm_sec);
  return ConfigOld_SetValue(root,
			 mode,
			 path,
			 buffer);
}


int ConfigOld_AddValue(CONFIGOLDGROUP *root,
		    unsigned int mode,
		    const char *path,
		    const char *value){
  CONFIGOLDVARIABLE *var;
  CONFIGOLDVALUE *val;

  assert(root);
  assert(path);
  DBG_VERBOUS(0, "AddValue for \"%s\"",path);
  var=(CONFIGOLDVARIABLE*)ConfigOld__GetPath(root,path,
				       mode | CONFIGOLDMODE_VARIABLE);
  if (!var) {
    DBG_DEBUG(0, "Path not available");
    return 1;
  }
  val=ConfigOld__Value_new(value);
  ConfigOld__AddValue(var, val);
  return 0;
}


int ConfigOld_ClearVariable(CONFIGOLDGROUP *root,
			 unsigned int mode,
			 const char *path){
  CONFIGOLDVARIABLE *var;

  assert(root);
  assert(path);
  DBG_VERBOUS(0, "ClearValue for \"%s\"",path);
  var=(CONFIGOLDVARIABLE*)ConfigOld__GetPath(root,path,
				       mode | CONFIGOLDMODE_VARIABLE);
  if (!var) {
    DBG_DEBUG(0, "Path not available");
    return 1;
  }
  ConfigOld__Variable_Clear(var);
  return 0;
}



CONFIGOLDGROUP *ConfigOld_new(){
  return ConfigOld__Group_new(0);
}


void ConfigOld_free(CONFIGOLDGROUP *g){
  ConfigOld__Group_free(g);
}


CONFIGOLDGROUP *ConfigOld_ParseLine(CONFIGOLDGROUP *root,
			      CONFIGOLDGROUP *group,
			      const char *s,
			      int mode) {
  CONFIGOLDVARIABLE *var;
  char name[256];
  char *np;
  char *p;
  const char *g;
  int i;
  int quotes;
  int esc;
  int firstval;

  assert(s);
  name[0]=0;

  /* check for group definition */
  g=s;
  while(*g && (unsigned char)(*g)<33)
    g++;
  if (*g=='[') {
    if (mode & CONFIGOLDMODE_ALLOW_GROUPS) {
      /* ok, parse group name */
      CONFIGOLDGROUP *grp;

      s=g;
      s++;
      while(*s && (unsigned char)(*s)<33)
	s++;
      p=name;
      i=sizeof(name)-1;
      while ((unsigned char)(*s)>31 && i && *s!=']' && *s!='#') {
	*p=*s;
	p++;
	s++;
      } /* while */
      if (!i) {
	DBG_ERROR(0, "Groupname is too long (limit is %d chars)",sizeof(name)-1);
	return 0;
      }
      if (*s!=']') {
	DBG_ERROR(0, "\"]\" expected");
	return 0;
      }
      *p=0;
      DBG_VERBOUS(0, "Selecting group \"%s\"",name);
      grp=ConfigOld__GetPath(root, name, mode &~CONFIGOLDMODE_VARIABLE);
      if (!grp) {
	DBG_DEBUG(0, "Group \"%s\" is not available",name);
	return 0;
      }
      return grp;
    }
    else {
      DBG_ERROR(0, "Group definition not allowed");
      return 0;
    }
  }

  /* get name */
  if (mode & CONFIGOLDMODE_REMOVE_STARTING_BLANKS)
    while(*s && (unsigned char)(*s)<33)
      s++;
  i=sizeof(name)-1;
  p=name;
  while ((unsigned char)(*s)>31 && i-- && *s!='=' && *s!='#') {
    *p=*s;
    p++;
    s++;
  } /* while */
  if (!i) {
    DBG_ERROR(0, "Name is too long (limit is %d chars)",sizeof(name)-1);
    return 0;
  }
  *p=0;
  np=name;

  /* post process name */
  if (mode & CONFIGOLDMODE_REMOVE_TRAILING_BLANKS) {
    i=strlen(name)-1;
    while (i>=0) {
      if ((unsigned char)(name[i])<33)
	name[i]=0;
      else
        break;
      i--;
    }

    if (mode & CONFIGOLDMODE_REMOVE_QUOTES) {
      i=strlen(name);
      if (i>1) {
	if (name[i-1]=='"' &&
	    name[0]=='"') {
	  name[i-1]=0;
	  np++;
	}
      }
    }
  }
  if ((unsigned char)(*s)<31 || *s=='#') {
    DBG_VERBOUS(0, "Empty line");
    return group;
  }

  /* get equation mark */
  if (*s!='=') {
    DBG_ERROR(0, "\"=\" expected");
    return 0;
  }
  s++;

  if (strlen(np)==0) {
    DBG_ERROR(0, "Variable name must not be empty");
    return 0;
  }

  DBG_VERBOUS(0, "Creating variable \"%s\"",np);
  if (mode & CONFIGOLDMODE_ALLOW_PATH_IN_VARS)
    var=ConfigOld__GetPath(group,
			np,
			mode|CONFIGOLDMODE_VARIABLE);
  else {
    if (!(mode & CONFIGOLDMODE_NAMECREATE_VARIABLE)) {
      var=group->variables;
      while(var) {
	if (strcasecmp(var->name, np)==0)
	  break;
        var=var->next;
      } /* while */
    }
    else
      var=0;
    if (!var) {
      var=ConfigOld__Variable_new(np,0);
      ConfigOld__AddVariable(group, var);
    }
  }
  if (!var) {
    DBG_DEBUG(0, "Could not create variable \"%s\"",np);
    return 0;
  }

  firstval=1;
  /* read komma separated values */
  while ((unsigned char)(*s)>31) {
    CONFIGOLDVALUE *val;
    char value[1024];
    char *vp;

    value[0]=0;

    /* skip komma that may occur */
    if (mode & CONFIGOLDMODE_REMOVE_STARTING_BLANKS)
      while(*s && (unsigned char)(*s)<33)
	s++;
    if (*s==0) {
      break;
    }
    if (*s==',') {
      if (firstval) {
	DBG_ERROR(0, "Unexpected komma");
	return 0;
      }
      s++;
    }
    else {
      if (!firstval) {
	DBG_ERROR(0, "Komma expected");
	return 0;
      }
    }

    /* get value */
    if (mode & CONFIGOLDMODE_REMOVE_STARTING_BLANKS)
      while(*s && (unsigned char)(*s)<33)
	s++;
    i=sizeof(value)-1;
    p=value;
    /* copy value */
    quotes=0;
    esc=0;
    while ((unsigned char)(*s)>31 && i) {
      if (esc) {
	*p=*s;
	p++;
        i--;
	esc=0;
      }
      else {
	if (*s=='\\')
	  esc=1;
	else if (*s=='"') {
	  quotes++;
	  if (quotes==2) {
	    s++;
	    break;
	  }
	}
	else if (*s=='#' && !(quotes&1))
	  break;
	else if (*s==',' && !(quotes&1))
	  break;
	else {
	  *p=*s;
	  p++;
	  i--;
	}
      }
      s++;
    } /* while */
    if (!i) {
      DBG_ERROR(0, "Value is too long (limit is %d chars)",sizeof(value)-1);
      return 0;
    }
    if (quotes&1) {
      DBG_ERROR(0, "Unbalanced quotation marks");
      return 0;
    }
    if (esc)
      DBG_WARN(0, "Backslash at the end of the line");
    *p=0;
    vp=value;
    /* post process value */
    if (mode & CONFIGOLDMODE_REMOVE_TRAILING_BLANKS && quotes==0) {
      i=strlen(value)-1;
      while (i>=0) {
	if ((unsigned char)(value[i])<33)
	  value[i]=0;
	else
          break;
	i--;
      }

      if (mode & CONFIGOLDMODE_REMOVE_QUOTES) {
	i=strlen(value);
	if (i>1) {
	  if (value[i-1]=='"' &&
	      value[0]=='"') {
	    value[i-1]=0;
	    vp++;
	  }
	}
      }
    }
    /* create value, append it */
    DBG_VERBOUS(0, " Creating value \"%s\"",vp);
    val=ConfigOld__Value_new(vp);
    ConfigOld__AddValue(var, val);

    if (*s=='#')
      break;
    firstval=0;
  } /* while (reading values) */

  return group;
}


int ConfigOld_ReadFile(CONFIGOLDGROUP *root,
		    const char *fname,
		    int mode){
  FILE *f;
  char lbuffer[2048];
  CONFIGOLDGROUP *curr;
  int ln;

  assert(root);
  assert(fname);


  curr=root;
  if (fname==0)
    f=stdin;
  else {
    f=fopen(fname,"r");
    if (f==0) {
      DBG_ERROR(0, "Error on fopen(%s): %s",fname,strerror(errno));
      return 1;
    }
  }

  ln=1;
  while (!feof(f)) {
    lbuffer[0]=0;
    if (fgets(lbuffer, sizeof(lbuffer)-1, f)==0)
      break;
    curr=ConfigOld_ParseLine(root, curr, lbuffer, mode);
    if (!curr) {
      DBG_ERROR(0, "Error in %s:%d",fname, ln);
      fclose(f);
      return 1;
    }
    ln++;
  } /* while */
  if (ferror(f)) {
    DBG_ERROR(0, "Error on fgets(%s)",fname);
    fclose(f);
    return 0;
  }
  if (fname)
    if (fclose(f)) {
      DBG_ERROR(0, "Error on fclose(%s): %s",fname,strerror(errno));
    }
  return 0;
}


int ConfigOld_ReadStream(CONFIGOLDGROUP *root,
                         GWEN_BUFFEREDIO *bio,
                         int mode){
  char lbuffer[2048];
  CONFIGOLDGROUP *curr;
  int ln;
  GWEN_ERRORCODE gerr;

  assert(root);

  curr=root;
  ln=1;
  gerr=0;
  while (!GWEN_BufferedIO_CheckEOF(bio)) {
    lbuffer[0]=0;
    gerr=GWEN_BufferedIO_ReadLine(bio, lbuffer, sizeof(lbuffer)-1);
    if (!GWEN_Error_IsOk(gerr)) {
      DBG_ERROR_ERR(0, gerr);
      return 1;
    }
    curr=ConfigOld_ParseLine(root, curr, lbuffer, mode);
    if (!curr) {
      DBG_ERROR(0, "Error in input stream (line %d)", ln);
      return 1;
    }
    ln++;
  } /* while */
  return 0;
}






int ConfigOld__WriteGroup(FILE *f, CONFIGOLDGROUP *root,
		       const char *path,
		       int mode) {
  CONFIGOLDGROUP *grp;
  CONFIGOLDVARIABLE *var;
  CONFIGOLDVALUE *val;
  int i;

  assert(root);

  DBG_VERBOUS(0, "Writing group under \"%s\"",path);
  /* write variables */
  var=root->variables;
  while (var) {
    int nameWritten;
    int lineSize;

    lineSize=0;
    nameWritten=0;

    val=var->values;
    i=0;
    while(val) {
      if (val->value) {
	if (*(val->value) || mode & CONFIGOLDMODE_EMPTY_VARIABLES) {
	  lineSize+=strlen(val->value)+1;
	  if (lineSize>CONFIGOLD_MAX_LINEWIDTH &&
	      (mode & CONFIGOLDMODE_BREAKLINES) && i) {
	    nameWritten=0;
	    lineSize=0;
	    i=0;
	    fprintf(f,"\n");
	  }
	  if (!nameWritten) {
	    fprintf(f,"%s=",var->name);
	    lineSize+=strlen(var->name)+1;
	    nameWritten=1;
	  }
	  if (i++) {
	    fprintf(f,",");
            lineSize++;
	  }
	  fprintf(f,"\"%s\"",val->value);
	  lineSize+=strlen(val->value)+2;
	}
	else {
	  fprintf(f,"\"\"");
	  lineSize+=2;
	}
      }
      val=val->next;
    } /* while */
    fprintf(f,"\n");
    var=var->next;
  } /* while */

  grp=root->groups;
  while(grp) {
    char pbuffer[256];
    int pblen;

    if (!grp->name) {
      DBG_ERROR(0, "Unnamed group");
      return 1;
    }

    pblen=strlen(grp->name)+1;
    if (path)
      pblen+=strlen(path);
    if (pblen>sizeof(pbuffer)-1) {
      DBG_ERROR(0, "Path too long (limit is %d bytes)",sizeof(pbuffer)-1);
      return 1;
    }
    /* print group name */
    pbuffer[0]=0;
    if (path) {
      if (strlen(path)) {
	strcpy(pbuffer,path);
	strcat(pbuffer,"/");
      }
    }
    strcat(pbuffer, grp->name);

    if (grp->variables ||
	(mode & CONFIGOLDMODE_EMPTY_GROUPS)) {
      fprintf(f,"\n[%s]\n",pbuffer);
    }
    DBG_VERBOUS(0, "About to write group \"%s\"",pbuffer);
    i=ConfigOld__WriteGroup(f, grp, pbuffer,mode);
    if (i) {
      DBG_ERROR(0, "Error writing group \"%s\"",grp->name);
      return i;
    }
    grp=grp->next;
  } /* while */
  return 0;
}


int ConfigOld_WriteFile(CONFIGOLDGROUP *root,
		     const char *fname,
		     int mode){
  FILE *f;
  int i;

  assert(root);
  assert(fname);

  if (fname==0)
    f=stdout;
  else {
    f=fopen(fname,"w+");
    if (f==0) {
      DBG_ERROR(0, "Error on fopen(%s): %s",fname,strerror(errno));
      return 1;
    }
  }
  i=ConfigOld__WriteGroup(f, root,root->name, mode);
  if (i) {
    DBG_ERROR(0, "Error writing group \"%s\"",root->name);
    fclose(f);
    return 1;
  }
  if (fname)
    if (fclose(f)) {
      DBG_ERROR(0, "Error on fclose(%s): %s",fname,strerror(errno));
    }

  return 0;
}


int ConfigOld_DumpGroup(FILE *f, CONFIGOLDGROUP *root,
		     int ins) {
  CONFIGOLDGROUP *grp;
  CONFIGOLDVARIABLE *var;
  CONFIGOLDVALUE *val;
  int i;

  assert(root);

  /* write variables */
  var=root->variables;
  while (var) {
    for (i=0; i<ins; i++)
      fprintf(f, "  ");
    fprintf(f,"%s=",var->name);
    val=var->values;
    i=0;
    while(val) {
      if (i++)
	fprintf(f,",");
      fprintf(f,"\"%s\"",val->value);
      val=val->next;
    } /* while */
    fprintf(f,"\n");
    var=var->next;
  } /* while */

  grp=root->groups;
  while(grp) {

    if (!grp->name) {
      DBG_ERROR(0, "Unnamed group");
      return 1;
    }

    for (i=0; i<ins; i++)
      fprintf(f, "  ");
    fprintf(f,"[%s]\n",grp->name);
    i=ConfigOld_DumpGroup(f, grp, ins+1);
    if (i) {
      DBG_ERROR(0, "Error writing group \"%s\"",grp->name);
      return i;
    }
    //fprintf(f,"\n");
    grp=grp->next;
  } /* while */
  return 0;
}



CONFIGOLDGROUP *ConfigOld_GetGroup(CONFIGOLDGROUP *root,
			     const char *path,
			     unsigned int mode){
  DBG_VERBOUS(0, "GetGroup for \"%s\"",path);
  return (CONFIGOLDGROUP*)ConfigOld__GetPath(root,
				       path,
				       mode & ~CONFIGOLDMODE_VARIABLE);
}


CONFIGOLDGROUP *ConfigOld_Duplicate(CONFIGOLDGROUP *root) {
  return ConfigOld__Group_duplicate(root);
}


int ConfigOld_Compare(const char *s1, const char *s2){
  if (s1==s2)
    return 0;
  if (s1==0 || s2==0)
    return 1;

  while(*s1 && *s2) {
    if (toupper(*s1)!=toupper(*s2))
      return 1;
    s1++;
    s2++;
  } /* while */
  return (*s1!=*s2);
}


void ConfigOld_ClearGroup(CONFIGOLDGROUP *g){
  assert(g);
  ConfigOld__Group_Clear(g);
}


void ConfigOld_AddVariable(CONFIGOLDGROUP *p, CONFIGOLDVARIABLE *v){
  ConfigOld__AddVariable(p, v);
}


void ConfigOld__ImportGroup(CONFIGOLDGROUP *root, CONFIGOLDGROUP *g) {
  CONFIGOLDGROUP *tmp;
  CONFIGOLDGROUP *gcopy;
  CONFIGOLDVARIABLE *vcopy;
  CONFIGOLDVARIABLE *currv;

  while(g) {
    DBG_DEBUG(0, "Importing group %s below %s",g->name, root->name);
    tmp=ConfigOld_GetGroup(root, g->name, CONFIGOLDMODE_NAMEMUSTEXIST);
    if (tmp==0) {
      /* does not exists, so simply add it and return */
      gcopy=ConfigOld_Duplicate(g);
      DBG_DEBUG(0, "Group \"%s\" does not exist, creating it", g->name);
      ConfigOld_AddGroup(root, gcopy);
    }
    else {
      /* group does exist, so import it */
      if (g->groups!=0) {
	ConfigOld__ImportGroup(tmp, g->groups);
	DBG_DEBUG(0, "Import of \"%s\" done.", g->name);
      }

      /* copy all variables */
      currv=g->variables;
      while (currv) {
	vcopy=ConfigOld__Variable_duplicate(currv);
	ConfigOld__AddVariable(tmp, vcopy);
	currv=currv->next;
      }
    }

    /* next group */
    g=g->next;
  } /* while */
}


void ConfigOld_ImportGroup(CONFIGOLDGROUP *root, CONFIGOLDGROUP *g) {
  ConfigOld__ImportGroup(root, g->groups);
}


int ConfigOld_SetBinValue(CONFIGOLDGROUP *root,
		       unsigned int mode,
		       const char *path,
		       const char *value,
		       unsigned int size){
  char *buffer;
  unsigned int l;

  l=(size*2)+1;
  buffer=(char*)malloc(l);
  assert(buffer);

  if (GWEN_Text_ToHex(value, size,
                      buffer, l)==0) {
    DBG_INFO(0, "From here");
    free(buffer);
    return -1;
  }
  ConfigOld_SetValue(root,
                     mode,
                     path, buffer);
  free(buffer);
  return 0;
}


int ConfigOld_GetBinValue(CONFIGOLDGROUP *root,
		       const char *path,
		       char *buffer,
		       unsigned int size,
		       int idx){
  const char *p;
  int rv;

  p=ConfigOld_GetValue(root, path, 0, idx);
  if (!p)
    return 0;
  rv=GWEN_Text_FromHex(p, buffer, size);
  if (rv<0) {
    DBG_INFO(0, "From here");
    return -1;
  }

  return rv;
}







