/***************************************************************************
 $RCSfile: oldconf.h,v $
                             -------------------
    cvs         : $Id: oldconf.h,v 1.3 2004/01/25 18:51:29 aquamaniac Exp $
    begin       : Mon Dec 02 2002
    copyright   : (C) 2002 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/

/**
 * @file chameleon/conf.h
 *
 * This file contains the configuration file parser and generator
 */

#ifndef CHAMELEON_CONF_H
#define CHAMELEON_CONF_H

#include <stdio.h>
#include <time.h>

#include <gwenhywfar/bufferedio.h>


#ifdef __cplusplus
extern "C" {
#endif

#define CONFIGOLD_MAX_LINEWIDTH 80

#define CONFIGOLDMODE_PATHMUSTEXIST          0x000001
#define CONFIGOLDMODE_PATHMUSTNOTEXIST       0x000002
#define CONFIGOLDMODE_PATHCREATE             0x000004

#define CONFIGOLDMODE_NAMEMUSTEXIST          0x000008
#define CONFIGOLDMODE_NAMEMUSTNOTEXIST       0x000010
#define CONFIGOLDMODE_NAMECREATE_GROUP       0x000020
#define CONFIGOLDMODE_NAMECREATE_VARIABLE    0x000040

#define CONFIGOLDMODE_VARIABLE               0x000080
#define CONFIGOLDMODE_OVERWRITE_VARS         0x000100
#define CONFIGOLDMODE_OVERWRITE_GROUPS       0x000200
#define CONFIGOLDMODE_REMOVE_QUOTES          0x000400
#define CONFIGOLDMODE_REMOVE_STARTING_BLANKS 0x000800
#define CONFIGOLDMODE_REMOVE_TRAILING_BLANKS 0x001000
#define CONFIGOLDMODE_ALLOW_PATH_IN_VARS     0x002000
#define CONFIGOLDMODE_ALLOW_GROUPS           0x004000
#define CONFIGOLDMODE_EMPTY_GROUPS           0x008000
#define CONFIGOLDMODE_EMPTY_VARIABLES        0x010000

  /**
   * If while writing a variable and a line gets longer than 80 and
   * there are still values left the variable name will be repeated
   * in the next line and the remaining values are printed there.
   */
#define CONFIGOLDMODE_BREAKLINES             0x020000

#define CONFIGOLDMODE_DEFAULT (\
  CONFIGOLDMODE_REMOVE_QUOTES | \
  CONFIGOLDMODE_REMOVE_STARTING_BLANKS | \
  CONFIGOLDMODE_REMOVE_TRAILING_BLANKS | \
  CONFIGOLDMODE_ALLOW_GROUPS | \
  CONFIGOLDMODE_BREAKLINES \
  )

typedef struct CONFIGOLDVALUESTRUCT CONFIGOLDVALUE;
typedef struct CONFIGOLDVARIABLESTRUCT CONFIGOLDVARIABLE;
typedef struct CONFIGOLDGROUPSTRUCT CONFIGOLDGROUP;


/**
 */
struct CONFIGOLDVALUESTRUCT {
  CONFIGOLDVALUE *next;
  char *value;
};


/**
 */
struct CONFIGOLDVARIABLESTRUCT {
  CONFIGOLDVARIABLE *next;
  char *name;
  CONFIGOLDGROUP *parent;
  CONFIGOLDVALUE *values;
};


/**
 */
struct CONFIGOLDGROUPSTRUCT {
  CONFIGOLDGROUP *next;
  char *name;
  CONFIGOLDGROUP *parent;
  CONFIGOLDGROUP *groups;
  CONFIGOLDVARIABLE *variables;
};


/**
 * Constructor. Please always use this constructor to create a configuration
 * structure.
 */
CONFIGOLDGROUP *ConfigOld_new();

/**
 * Destructor.
 */
void ConfigOld_free(CONFIGOLDGROUP *g);


/**
 * Get the value of a variable or a default value if the variable does not
 * exist or doesn't have a value.
 * @param root root to which the path is relative
 * @param path path and name of the variable (e.g. "global/settings/var1")
 * @param defaultValue this value is returned if the variable does not exist
 * or if the variable has no value
 * @param idx a variable may have multiple values, so you can choose here
 * which value you want (starting with 0)
 */
const char *ConfigOld_GetValue(CONFIGOLDGROUP *root,
					  const char *path,
					  const char *defaultValue,
					  int idx);

/**
 * Get an integer value. This internally calls @ref ConfigOld_GetValue, please
 * see there for details.
 */
int ConfigOld_GetIntValue(CONFIGOLDGROUP *root,
				    const char *path,
				    int defaultValue,
				    int idx);


/**
 * Get a binary value. This internally calls @ref ConfigOld_GetValue, please
 * see there for details.
 * @return number of bytes copied into the buffer (-1 on error)
 */
int ConfigOld_GetBinValue(CONFIGOLDGROUP *root,
				     const char *path,
				     char *buffer,
				     unsigned int size,
				     int idx);

/**
 * Get a boolean value. This internally calls @ref ConfigOld_GetValue, please
 * see there for details.
 */
int ConfigOld_GetBoolValue(CONFIGOLDGROUP *root,
				      const char *path,
				      int defaultValue,
				      int idx);

/**
 * Get a time value. Format of the parsed string is:
 * "YEAR/MONTH/DAY-HOUR:MINUTE:SECOND"
 */
time_t ConfigOld_GetTimeValue(CONFIGOLDGROUP *root,
			   const char *path,
			   time_t defaultValue,
			   int idx);

/**
 * Set the value of a variable in respect to the mode given.
 * @param root root to which the path is relative
 * @param mode this specifies how the value is set. This mode consists
 * of one or multiple "CONFIGOLDMODE_XXX" values which may be ORed.
 * @param path path of the variable (e.g. "global/settings/var1")
 * @param value value to set
 */
int ConfigOld_SetValue(CONFIGOLDGROUP *root,
				 unsigned int mode,
				 const char *path,
				 const char *value);

/**
 * This is just a conveniance function which internally calls
 * @ref ConfigOld_SetValue. Please see there for details.
 */
int ConfigOld_SetIntValue(CONFIGOLDGROUP *root,
				    unsigned int mode,
				    const char *path,
				    int value);

int ConfigOld_SetBinValue(CONFIGOLDGROUP *root,
				     unsigned int mode,
				     const char *path,
				     const char *value,
				     unsigned size);


/**
 * This is just a conveniance function which internally calls
 * @ref ConfigOld_SetValue. Please see there for details.
 */
int ConfigOld_SetBinValue(CONFIGOLDGROUP *root,
				     unsigned int mode,
				     const char *path,
				     const char *value,
				     unsigned int size);

/**
 * This is just a conveniance function which internally calls
 * @ref ConfigOld_SetValue. Please see there for details.
 */
int ConfigOld_SetBoolValue(CONFIGOLDGROUP *root,
				      unsigned int mode,
				      const char *path,
				      int value);

int ConfigOld_SetTimeValue(CONFIGOLDGROUP *root,
			unsigned int mode,
			const char *path,
			time_t value);

/**
 * Adds a value to a given variable.
 * @param root root to which the path is relative
 * @param mode this specifies how the value is set. This mode consists
 * of one or multiple "CONFIGOLDMODE_XXX" values which may be ORed.
 * @param path path of the variable (e.g. "global/settings/var1")
 * @param value value to add
 */
int ConfigOld_AddValue(CONFIGOLDGROUP *root,
				 unsigned int mode,
				 const char *path,
				 const char *value);

/**
 * Clears a variable, so all it's values are destroyed.
 */
int ConfigOld_ClearVariable(CONFIGOLDGROUP *root,
				      unsigned int mode,
				      const char *path);

/**
 * This function parses a text line and stores variables/groups read
 * in the given root.
 * @return current group (this may have changed depending on the content
 * of the text line just parsed
 * @param root root to which all paths are relative
 * @param group current group within the root, this is the group all variables
 * parsed are relative to. Please note that group paths are relative to
 * the root, not to group !
 * @param mode specifies how to store the groups and variables. The mode
 * consists of one or multiple CONFIGOLDMODE_XYZ values which may be logically
 * ORed.
 */
CONFIGOLDGROUP *ConfigOld_ParseLine(CONFIGOLDGROUP *root,
					   CONFIGOLDGROUP *group,
					   const char *s,
					   int mode);

/**
 * Reads a configuration file and parses it.
 * @param fname path and name of the file to read.
 * @param mode specifies how to store the groups and variables. The mode
 * consists of one or multiple CONFIGOLDMODE_XYZ values which may be logically
 * ORed.
 */
int ConfigOld_ReadFile(CONFIGOLDGROUP *root,
				 const char *fname,
				 int mode);

/**
 * Creates a configuration file from the given configuration.
 * @param fname path and name of the file to create
 * @param mode specifies how to store the groups and variables. The mode
 * consists of one or multiple CONFIGOLDMODE_XYZ values which may be logically
 * ORed.
 */
int ConfigOld_WriteFile(CONFIGOLDGROUP *root,
				  const char *fname,
				  int mode);

/**
 * Adds a variable to a given group.
 */
void ConfigOld_AddVariable(CONFIGOLDGROUP *p, CONFIGOLDVARIABLE *v);


/**
 * Dumps the content of a configuration into an open file.
 * This is used for debugging purposes, where f is stderr.
 * @param f file to dump to (most cases "stderr")
 * @param root group to dump
 * @param ins number of spaces to prepend to each output line, this is used
 * to show hierarchies.
 */
int ConfigOld_DumpGroup(FILE *f, CONFIGOLDGROUP *root,
				  int ins);

/**
 * Lookup a group by its name.
 */
CONFIGOLDGROUP *ConfigOld_GetGroup(CONFIGOLDGROUP *root,
					  const char *path,
					  unsigned int mode);


/**
 * Adds a group to another one. The new group becomes the child of the
 * existing one.
 * @param p parent (group to which to add the new group)
 * @param w new group to add
 */
void ConfigOld_AddGroup(CONFIGOLDGROUP *p, CONFIGOLDGROUP *w);

/**
 * Adds all groups and variables which are the children of the given
 * group to the given parent by making deep copies.
 * The difference between this an ConfigOld_ImportGroup is, that
 * this function imports ALL children of the given group, while the other
 * function only imports the groups from below the given one.
 */
void ConfigOld_ImportTreeChildren(CONFIGOLDGROUP *p, CONFIGOLDGROUP *g);

/**
 * Unlinks a group from the configuration. This group will not be destroyed.
 * @todo Hmm, this function may change, since the parent pointer is already
 * stored within the group struct.
 * @param p parent of the group
 */
void ConfigOld_UnlinkGroup(CONFIGOLDGROUP *p, CONFIGOLDGROUP *w);

/**
 * Clears a group, so all it's subgroups and variables are destroyed.
 */
void ConfigOld_ClearGroup(CONFIGOLDGROUP *g);


/**
 * Duplicates a group. The new group is a deep copy of the original.
 */
CONFIGOLDGROUP *ConfigOld_Duplicate(CONFIGOLDGROUP *root);

/**
 * Imports a new group into an existing configuration (including all the
 * subgroups and variables of the given group)
 */
void ConfigOld_ImportGroup(CONFIGOLDGROUP *root, CONFIGOLDGROUP *g);


/**
 * Just a conveniance function for case insensitive compares.
 * Well, meanwhile I now that there already are such functions ;-)
 * @return result of "s1-s2"
 */
int ConfigOld_Compare(const char *s1, const char *s2);


int ConfigOld_ReadStream(CONFIGOLDGROUP *root,
                         GWEN_BUFFEREDIO *bio,
                         int mode);

#ifdef __cplusplus
}
#endif


#endif




