/***************************************************************************
  $RCSfile: storagefile.cpp,v $
  -------------------
  cvs         : $Id: storagefile.cpp,v 1.6 2004/01/25 18:51:30 aquamaniac Exp $
  begin       : Mon Oct 13 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "storagefile.h"
#include <gwenhywfar/debug.h>
#include <gwenhywfar/directory.h>
#include <gwenhywfar/path.h>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>



TransactionStorageFile::OpenDay::OpenDay(int year, int month, int day,
                                         const string &path,
                                         unsigned int hdl,
                                         GWEN_DB_NODE *data,
                                         bool readonly)
:_year(year)
,_month(month)
,_day(day)
,_path(path)
,_handle(hdl)
,_data(data)
,_readOnly(readonly)
,_changed(false)
,_currTrans(0){

}


TransactionStorageFile::OpenDay::~OpenDay(){
  GWEN_DB_Group_free(_data);
}


GWEN_DB_NODE *TransactionStorageFile::OpenDay::getFirstTransaction(){
  GWEN_DB_NODE *n;

  n=GWEN_DB_GetGroup(_data,
                     GWEN_DB_FLAGS_DEFAULT | GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                     "transactions");
  if (!n) {
    DBG_INFO(0, "No transactions");
    return 0;
  }
  n=GWEN_DB_GetFirstGroup(n);
  _currTrans=n;
  return n;
}



GWEN_DB_NODE *TransactionStorageFile::OpenDay::getNextTransaction(){
  GWEN_DB_NODE *n;

  if (!_currTrans)
    return 0;

  n=GWEN_DB_GetNextGroup(_currTrans);
  _currTrans=n;
  return n;
}




unsigned int TransactionStorageFile::_nextHandle=1;



TransactionStorageFile::TransactionStorageFile(AqMoneyAPI *api,
                                               const string &dir,
                                               const string &subdir,
                                               int country,
                                               const string &bankCode,
                                               const string &account)
:TransactionStorage(api, country, bankCode, account)
, _dir(dir)
, _subdir(subdir)
, _connected(false){

}


TransactionStorageFile::TransactionStorageFile(AqMoneyAPI *api,
                                               GWEN_DB_NODE *db,
                                               int country,
                                               const string &bc,
                                               const string &ac) throw (Error)
:TransactionStorage(api, db, country, bc, ac)
, _connected(false){
  Error err;

  err=fromConfig(db);
  if (!err.isOk())
    throw Error("TransactionStorageFile::TransactionStorageFile", err);
}




TransactionStorageFile::~TransactionStorageFile(){

}




void *TransactionStorageFile::_handlePathElement(const char *entry,
                                                 void *data,
                                                 unsigned int flags){
  char *p;
  struct stat st;
  bool exists;

  p=(char*)data;
  if ((strlen(p)+strlen(entry)+2)>=256) {
    DBG_ERROR(0, "Buffer too small");
    return 0;
  }
  strcat(p, "/");
  strcat(p, entry);

  // check for existence of the file/folder

  DBG_DEBUG(0, "Checking entry \"%s\"", p);
  if (stat(p, &st)) {
    exists=false;
    DBG_DEBUG(0, "stat: %s (%s)", strerror(errno), p);
    if ((flags & GWEN_PATH_FLAGS_PATHMUSTEXIST) ||
        ((flags & GWEN_PATH_FLAGS_LAST) &&
         (flags & GWEN_PATH_FLAGS_NAMEMUSTEXIST))) {
      DBG_ERROR(0, "Path \"%s\" does not exist (it should)", p);
      return 0;
    }
  }
  else {
    DBG_DEBUG(0, "Checking for type");
    exists=true;
    if (flags & GWEN_PATH_FLAGS_VARIABLE) {
      if (!S_ISREG(st.st_mode)) {
        DBG_ERROR(0, "%s not a regular file", p);
        return 0;
      }
    }
    else {
      if (!S_ISDIR(st.st_mode)) {
        DBG_ERROR(0, "%s not a direcory", p);
        return 0;
      }
    }
    if ((flags & GWEN_PATH_FLAGS_PATHMUSTNOTEXIST) ||
        ((flags & GWEN_PATH_FLAGS_LAST) &&
         (flags & GWEN_PATH_FLAGS_NAMEMUSTNOTEXIST))) {
      DBG_ERROR(0, "Path \"%s\" does not exist (it should)", p);
      return 0;
    }
  } // if stat is ok

  if (!exists) {
    DBG_DEBUG(0, "Entry \"%s\" does not exist", p);
    if (flags & GWEN_PATH_FLAGS_VARIABLE) {
      // create file
      int fd;

      DBG_DEBUG(0, "Creating file \"%s\"", p);
      fd=open(p, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
      if (fd==-1) {
        DBG_ERROR(0, "open: %s (%s)", strerror(errno), p);
        return 0;
      }
      close(fd);
      DBG_DEBUG(0, "Sucessfully created");
    }
    else {
      // create dir
      DBG_DEBUG(0, "Creating folder \"%s\"", p);

      if (mkdir(p, S_IRWXU)) {
        DBG_ERROR(0, "mkdir: %s (%s)", strerror(errno), p);
        return 0;
      }
    }
  } // if exists
  else {
    DBG_DEBUG(0, "Entry \"%s\" exists", p);
  }
  DBG_DEBUG(0, "Returning this: %s", p);
  return p;
}







string TransactionStorageFile::_getPath(int year, int month, int day,
                                        bool create) {
  string dname;
  char buffer[256];
  void *p;
  unsigned int flags;

  flags=GWEN_PATH_FLAGS_VARIABLE;
  if (!create)
    flags|=GWEN_PATH_FLAGS_NAMEMUSTEXIST;

  dname=api()->applicationDataDir();
  dname+="/";
  dname+=String::num2string(country(), true, 3);
  dname+="/";
  dname+=bankCode();
  dname+="/accounts/";
  dname+=account();
  if (!_dir.empty()) {
    dname+="/";
    dname+=_dir;
  }
  if (!_subdir.empty()) {
    dname+="/";
    dname+=_subdir;
  }
  dname+="/";
  dname+=String::num2string(year, true, 4);
  dname+="/";
  dname+=String::num2string(month, true, 2);
  dname+="/";
  dname+=String::num2string(day, true, 2);
  dname+=".trans";

  DBG_INFO(0, "Searching in \"%s\"", dname.c_str());
  if (dname.length()>=sizeof(buffer)) {
    DBG_ERROR(0, "Path too long");
    return "";
  }

  buffer[0]=0;
  p=buffer;
  p=GWEN_Path_Handle(dname.c_str(), p,
                     GWEN_PATH_FLAGS_VARIABLE,
                     _handlePathElement);
  if (p==0)
    return "";
  return (const char*)p;
}



unsigned int TransactionStorageFile::openDay(int year,
                                             int month,
                                             int day,
                                             bool write) throw(Error){
  string path;
  GWEN_DB_NODE *db;
  GWEN_DB_NODE *n;

  path=_getPath(year, month, day, write);
  if (path.empty()) {
    throw Error("TransactionStorageFile::openDay",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Could not open day file");
  }
  db=GWEN_DB_Group_new("dayfile");
  // read file
  DBG_INFO(0, "Reading file \"%s\"", path.c_str());
  if (GWEN_DB_ReadFile(db, path.c_str(),
                       GWEN_DB_FLAGS_DEFAULT|
                       GWEN_PATH_FLAGS_CREATE_GROUP)) {
    throw Error("TransactionStorageFile::openDay",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Could not read day file");
  }

  // TODO: Remove this after some time
  n=GWEN_DB_GetGroup(db,
                     GWEN_DB_FLAGS_DEFAULT |
                     GWEN_PATH_FLAGS_NAMEMUSTEXIST,
                     "transactions");
  if (n) {
    n=GWEN_DB_GetFirstGroup(n);
    while(n) {
      if (strcasecmp(GWEN_DB_GroupName(n), "transaction")==0) {
        int tk;

        tk=GWEN_DB_GetIntValue(n, "code", 0, -1);
        if (tk!=-1) {
          GWEN_DB_DeleteVar(n, "code");
          GWEN_DB_SetCharValue(n,
                               GWEN_DB_FLAGS_DEFAULT |
                               GWEN_DB_FLAGS_OVERWRITE_VARS,
                               "textkey",
                               String::num2string(tk).c_str());
        }
      }
      n=GWEN_DB_GetNextGroup(n);
    }
  }
  // end of TODO

  _openDays.push_back(new OpenDay(year, month, day, path, _nextHandle, db,
                                  !write));
  DBG_INFO(0, "Day %d/%d/%d open (%d)",
           year, month, day, _nextHandle);
  if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
    GWEN_DB_Dump(db, stderr, 4);
  return _nextHandle++;
}


Pointer<TransactionStorageFile::OpenDay>
TransactionStorageFile::_findOpenDay(unsigned int hdl) {
  list<Pointer<OpenDay> >::iterator it;

  for (it=_openDays.begin();
       it!=_openDays.end();
       it++) {
    if ((*it).ref().handle()==hdl)
      return *it;
  } // for
  return 0;
}


Error TransactionStorageFile::clearDay(unsigned int hdl){
  Pointer<OpenDay> d;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    return Error("TransactionStorageFile::clearDay",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Open day not found");
  }
  GWEN_DB_ClearGroup(d.ref().data(),0);
  DBG_INFO(0, "Day %d/%d/%d in \"%s\" cleared",
           d.ref().year(), d.ref().month(), d.ref().day(), _subdir.c_str());
  d.ref().setChanged(true);
  return Error();
}



Error TransactionStorageFile::closeDay(unsigned int hdl){
  list<Pointer<OpenDay> >::iterator it;

  for (it=_openDays.begin();
       it!=_openDays.end();
       it++) {
    if ((*it).ref().handle()==hdl) {
      if ((*it).ref().changed()) {
        // save changes
        if (GWEN_DB_WriteFile((*it).ref().data(),
                              (*it).ref().path().c_str(),
                              GWEN_DB_FLAGS_DEFAULT)) {
          return Error("TransactionStorageFile::closeDay",
                       ERROR_LEVEL_NORMAL,
                       HBCI_ERROR_CODE_UNKNOWN,
                       ERROR_ADVISE_DONTKNOW,
                       "Could not write day file");
        }
      }
      _openDays.erase(it);
      return Error();
    }
  } // for
  return Error("TransactionStorageFile::closeDay",
               ERROR_LEVEL_NORMAL,
               HBCI_ERROR_CODE_UNKNOWN,
               ERROR_ADVISE_DONTKNOW,
               "Open day not found");
}



Error TransactionStorageFile::abandonDay(unsigned int hdl){
  list<Pointer<OpenDay> >::iterator it;

  for (it=_openDays.begin();
       it!=_openDays.end();
       it++) {
    if ((*it).ref().handle()==hdl) {
      DBG_INFO(0, "Abandoning changes to open day");
      _openDays.erase(it);
      return Error();
    }
  } // for
  return Error("TransactionStorageFile::abandonDay",
               ERROR_LEVEL_NORMAL,
               HBCI_ERROR_CODE_UNKNOWN,
               ERROR_ADVISE_DONTKNOW,
               "Open day not found");
}



Error TransactionStorageFile::addTransaction(unsigned int hdl,
                                             GWEN_DB_NODE *xa){
  GWEN_DB_NODE *n;
  Pointer<OpenDay> d;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    return Error("TransactionStorageFile::addTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Open day not found");
  }

  DBG_INFO(0, "Adding transaction for day %d/%d/%d",
           d.ref().year(), d.ref().month(), d.ref().day());
  if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
    GWEN_DB_Dump(d.ref().data(), stderr, 1);
  n=GWEN_DB_GetGroup(d.ref().data(),
                     GWEN_DB_FLAGS_DEFAULT |
                     GWEN_PATH_FLAGS_CREATE_GROUP,
                     "transactions/transaction");
  assert(n);
  if (GWEN_DB_AddGroupChildren(n, xa)) {
    return Error("TransactionStorageFile::addTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Could not add transaction");
  }
  DBG_INFO(0, "Added transaction for day %d/%d/%d",
           d.ref().year(), d.ref().month(), d.ref().day());
  d.ref().setChanged(true);
  return Error();
}



bool TransactionStorageFile::_cmpCharValue(GWEN_DB_NODE *xa,
                                           GWEN_DB_NODE *nn,
                                           const char *name,
                                           unsigned int idx){
  const char *p1, *p2;

  p1=GWEN_DB_GetCharValue(xa, name, idx, 0);
  p2=GWEN_DB_GetCharValue(nn, name, idx, 0);
  if (!p1 && !p2)
    return true;
  if (p1) {
    // given in first group, so we need to check for it
    if (!p2)
      // not given in second, so it matches
      return true;
    // otherwise compare
    return (strcasecmp(p1, p2)==0);
  }
  // not given in first, and it still is the first index, so it matches
  return (idx==0);
}


bool TransactionStorageFile::_cmpIntValue(GWEN_DB_NODE *xa,
                                          GWEN_DB_NODE *nn,
                                          const char *name,
                                          unsigned int idx){
  int p1, p2;

  p1=GWEN_DB_GetIntValue(xa, name, idx, -1);
  p2=GWEN_DB_GetIntValue(nn, name, idx, -1);
  return p1==p2;
}


GWEN_DB_NODE *TransactionStorageFile::_findTransaction(Pointer<OpenDay> d,
                                                       GWEN_DB_NODE *xa){
  GWEN_DB_NODE *n;
  GWEN_DB_NODE *nn;

  n=GWEN_DB_GetGroup(d.ref().data(),
                     GWEN_DB_FLAGS_DEFAULT |
                     GWEN_PATH_FLAGS_CREATE_GROUP,
                     "transactions");
  assert(n);
  nn=GWEN_DB_GetFirstGroup(n);
  while(nn) {
    if (strcasecmp(GWEN_DB_GroupName(nn), "transaction")==0) {
      const char *p1, *p2;
      unsigned int i;

      if (
          _cmpCharValue(xa, nn, "date", 0) &&
          _cmpCharValue(xa, nn, "value/value", 0) &&
          _cmpCharValue(xa, nn, "value/currency", 0) &&
          _cmpCharValue(xa, nn, "textkey", 0) &&
          _cmpCharValue(xa, nn, "jobid", 0) &&
          _cmpCharValue(xa, nn, "otheraccount/accountid", 0) &&
          _cmpIntValue(xa, nn, "otheraccount/country", 0) &&
          _cmpCharValue(xa, nn, "otheraccount/bankcode", 0) &&
          _cmpCharValue(xa, nn, "ouraccount/accountid", 0) &&
          _cmpIntValue(xa, nn, "ouraccount/country", 0) &&
          _cmpCharValue(xa, nn, "ouraccount/bankcode", 0)){
        bool lmatch;

        lmatch=true;

        /* compare othername */
        if (lmatch) {
          for (i=0; ; i++) {
            p1=GWEN_DB_GetCharValue(xa, "othername", i, 0);
            p2=GWEN_DB_GetCharValue(nn, "othername", i, 0);
            if (!p1 && !p2)
              break;
            if (!(p1 && p2)) {
              lmatch=false;
              break;
            }
            if (strcasecmp(p1, p2)!=0) {
            lmatch=false;
            break;
            }
          }
        }

        /* compare purpose lines */
        if (lmatch) {
          for (i=0; ; i++) {
            p1=GWEN_DB_GetCharValue(xa, "purpose", i, 0);
            p2=GWEN_DB_GetCharValue(nn, "purpose", i, 0);
            if (!(p1 && p2))
              break;
            if (strcasecmp(p1, p2)!=0)
              break;
          }
        }

        if (lmatch) {
          // still matches, so we found the transaction
          return nn;
        }
      }
    } /* if groupname matches */
    nn=GWEN_DB_GetNextGroup(nn);
  } /* while */

  return 0;
}



Error TransactionStorageFile::delTransaction(unsigned int hdl,
                                             GWEN_DB_NODE *xa){
  GWEN_DB_NODE *n;
  Pointer<OpenDay> d;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    return Error("TransactionStorageFile::delTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Open day not found");
  }

  n=_findTransaction(d, xa);
  if (n) {
    GWEN_DB_UnlinkGroup(n);
    GWEN_DB_Group_free(n);
    d.ref().setChanged(true);
    DBG_ERROR(0, "Deleted transaction for day %d/%d/%d",
              d.ref().year(), d.ref().month(), d.ref().day());
    return Error();
  }
  else
    return Error("TransactionStorageFile::delTransaction",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Transaction not found");
}



Error TransactionStorageFile::setOpeningBalance(unsigned int hdl,
                                                GWEN_DB_NODE *bal){
  GWEN_DB_NODE *n;
  Pointer<OpenDay> d;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    return Error("TransactionStorageFile::setOpeningBalance",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Open day not found");
  }

  n=GWEN_DB_GetGroup(d.ref().data(),
                     GWEN_DB_FLAGS_DEFAULT,
                     "openingBalance");
  assert(n);
  DBG_DEBUG(0, "Before opening balance:");
  if (GWEN_Logger_GetLevel(0)>=GWEN_LoggerLevelDebug)
    GWEN_DB_Dump(n, stderr, 1);
  if (GWEN_DB_AddGroupChildren(n, bal)) {
    return Error("TransactionStorageFile::setOpeningBalance",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Could not add transaction");
  }
  DBG_INFO(0, "Set opening balance for day %d/%d/%d",
           d.ref().year(), d.ref().month(), d.ref().day());
  d.ref().setChanged(true);
  return Error();
}



Error TransactionStorageFile::setClosingBalance(unsigned int hdl,
                                                GWEN_DB_NODE *bal){
  GWEN_DB_NODE *n;
  Pointer<OpenDay> d;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    return Error("TransactionStorageFile::setClosingBalance",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Open day not found");
  }

  n=GWEN_DB_GetGroup(d.ref().data(),
                     GWEN_DB_FLAGS_DEFAULT,
                     "closingBalance");
  assert(n);
  if (GWEN_DB_AddGroupChildren(n, bal)) {
    return Error("TransactionStorageFile::setOpeningBalance",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Could not add transaction");
  }
  DBG_INFO(0, "Set closing balance for day %d/%d/%d",
           d.ref().year(), d.ref().month(), d.ref().day());
  d.ref().setChanged(true);
  return Error();
}



GWEN_DB_NODE *
TransactionStorageFile::getFirstTransaction(unsigned int hdl){
  Pointer<OpenDay> d;
  GWEN_DB_NODE *n;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    throw Error("TransactionStorageFile::getFirstTransaction",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Open day not found");
  }

  n=d.ref().getFirstTransaction();
  if (n)
    return GWEN_DB_Group_dup(n);
  return 0;
}



GWEN_DB_NODE *
TransactionStorageFile::getNextTransaction(unsigned int hdl){
  Pointer<OpenDay> d;
  GWEN_DB_NODE *n;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    throw Error("TransactionStorageFile::getNextTransaction",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Open day not found");
  }

  n=d.ref().getNextTransaction();
  if (n)
    return GWEN_DB_Group_dup(n);
  return 0;
}



GWEN_DB_NODE *
TransactionStorageFile::getOpeningBalance(unsigned int hdl) throw(Error){
  Pointer<OpenDay> d;
  GWEN_DB_NODE *n;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    throw Error("TransactionStorageFile::getOpeningBalance",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Open day not found");
  }

  n=GWEN_DB_GetGroup(d.ref().data(),
                     GWEN_DB_FLAGS_DEFAULT,
                     "openingBalance");
  if (n)
    return GWEN_DB_Group_dup(n);
  return 0;
}



GWEN_DB_NODE *
TransactionStorageFile::getClosingBalance(unsigned int hdl) throw(Error){
  Pointer<OpenDay> d;
  GWEN_DB_NODE *n;

  d=_findOpenDay(hdl);
  if (!d.isValid()) {
    throw Error("TransactionStorageFile::getClosingBalance",
                ERROR_LEVEL_NORMAL,
                HBCI_ERROR_CODE_UNKNOWN,
                ERROR_ADVISE_DONTKNOW,
                "Open day not found");
  }

  n=GWEN_DB_GetGroup(d.ref().data(),
                     GWEN_DB_FLAGS_DEFAULT,
                     "closingBalance");
  if (n)
    return GWEN_DB_Group_dup(n);
  return 0;
}



Error TransactionStorageFile::_buildIndex() {
  GWEN_DIRECTORYDATA *dyears;
  string dname;

  dname=api()->applicationDataDir();
  dname+="/";
  dname+=String::num2string(country(), true, 3);
  dname+="/";
  dname+=bankCode();
  dname+="/accounts/";
  dname+=account();
  if (!_dir.empty()) {
    dname+="/";
    dname+=_dir;
  }
  if (!_subdir.empty()) {
    dname+="/";
    dname+=_subdir;
  }

  _availDays.clear();
  dyears=GWEN_Directory_new();
  if (!GWEN_Directory_Open(dyears, dname.c_str())) {
    char nbuffer[256];

    // check for years
    while(!GWEN_Directory_Read(dyears,
                               nbuffer,
                               sizeof(nbuffer))) {
      unsigned int year;

      if (1==sscanf(nbuffer, "%d", &year)) {
        GWEN_DIRECTORYDATA *dmonths;
        string monthpath;

        // found a year, check it further
        monthpath=dname+"/"+nbuffer;
        dmonths=GWEN_Directory_new();
        if (!GWEN_Directory_Open(dmonths, monthpath.c_str())) {

          // check for months
          while(!GWEN_Directory_Read(dmonths,
                                     nbuffer,
                                     sizeof(nbuffer))) {
            unsigned int month;

            if (1==sscanf(nbuffer, "%d", &month)) {
              GWEN_DIRECTORYDATA *ddays;
              string daypath;

              // found a month, check it further
              daypath=monthpath+"/"+nbuffer;
              ddays=GWEN_Directory_new();
              if (!GWEN_Directory_Open(ddays, daypath.c_str())) {
                while(!GWEN_Directory_Read(ddays,
                                           nbuffer,
                                           sizeof(nbuffer))) {
                  int i;

                  i=strlen(nbuffer);
                  if (i>6) {
                    if (strcmp(nbuffer+i-6, ".trans")==0) {
                      unsigned int day;

                      nbuffer[i-6]=0;
                      if (1==sscanf(nbuffer, "%d", &day)) {
                        // found a day, save it
                        DBG_INFO(0, "Found day %d/%d/%d (%08x)",
                                 year, month, day,
                                 (year*65536+month*256+day));
                        _availDays.push_back(year*65536+month*256+day);
                      } // if entry is a day
                    } // if filename ends with ".trans"
                  } // if filename is long enough
                } // while still days
                if (GWEN_Directory_Close(ddays)) {
                  return Error("TransactionStorageFile::_buildIndex",
                               ERROR_LEVEL_NORMAL,
                               HBCI_ERROR_CODE_UNKNOWN,
                               ERROR_ADVISE_DONTKNOW,
                               "Error closing dir",
                               daypath);
                }
              } // if open
              GWEN_Directory_free(ddays);
            } // if entry is a month
          } // while still months
          if (GWEN_Directory_Close(dmonths)) {
            return Error("TransactionStorageFile::_buildIndex",
                         ERROR_LEVEL_NORMAL,
                         HBCI_ERROR_CODE_UNKNOWN,
                         ERROR_ADVISE_DONTKNOW,
                         "Error closing dir",
                         monthpath);
          }
        } // if monthpath ok
        GWEN_Directory_free(dmonths);
      } // if entry is a year
    } /* while years */
    if (GWEN_Directory_Close(dyears)) {
      return Error("TransactionStorageFile::_buildIndex",
                   ERROR_LEVEL_NORMAL,
                   HBCI_ERROR_CODE_UNKNOWN,
                   ERROR_ADVISE_DONTKNOW,
                   "Error closing dir",
                   dname);
    }
  } // if open is ok
  GWEN_Directory_free(dyears);

  // sort available days
  _availDays.sort();

  return Error();
}



Error TransactionStorageFile::connect(){
  if (_connected)
    return Error("TransactionStorageFile::connect",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Already connected");
  _connected=true;
  return Error();
}



Error TransactionStorageFile::disconnect(){
  if (!_connected)
    return Error("TransactionStorageFile::disconnect",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Not connected");
  _availDays.clear();
  return Error();
}



bool TransactionStorageFile::connected(){
  return _connected;
}



Error TransactionStorageFile::getFirstDay(int &year,
                                          int &month,
                                          int &day){
  Error err;
  unsigned int d;

  DBG_INFO(0, "Getting first day");
  err=_buildIndex();
  if (!err.isOk())
    return Error("TransactionStorageFile::getFirstDay", err);
  _availDaysIt=_availDays.begin();
  if (_availDaysIt==_availDays.end()) {
    return Error("TransactionStorageFile::getFirstDay",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "No days stored");
  }
  d=*(_availDaysIt++);
  DBG_INFO(0, "Day is %08x", d);
  year=(d>>16) & 0xffff;
  month=(d>>8) & 0xff;
  day=d & 0xff;
  return Error();
}



Error TransactionStorageFile::getNextDay(int &year,
                                         int &month,
                                         int &day){
  unsigned int d;

  if (_availDaysIt==_availDays.end()) {
    return Error("TransactionStorageFile::getNextDay",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "No more days stored");
  }
  d=*(_availDaysIt++);
  year=(d>>16) & 0xffff;
  month=(d>>8) & 0xff;
  day=d & 0xff;
  return Error();
}



Error TransactionStorageFile::fromConfig(GWEN_DB_NODE *db){
  const char *p;

  p=GWEN_DB_GetCharValue(db, "type", 0, "");
  if (strcasecmp(p, STORAGEFILE_TYPENAME)!=0) {
    return Error("TransactionStorageFile::fromConfig",
                 ERROR_LEVEL_NORMAL,
                 HBCI_ERROR_CODE_UNKNOWN,
                 ERROR_ADVISE_DONTKNOW,
                 "Not my configuration",
                 "bad type");
  }
  _dir=GWEN_DB_GetCharValue(db, "dir", 0, "");
  _subdir=GWEN_DB_GetCharValue(db, "subdir", 0, "");
  DBG_INFO(0, "TransactionStorageFile successfully loaded");
  return Error();
}



Error TransactionStorageFile::toConfig(GWEN_DB_NODE *db){
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT,
                       "type",
                       STORAGEFILE_TYPENAME);
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT,
                       "dir",
                       _dir.c_str());
  GWEN_DB_SetCharValue(db, GWEN_DB_FLAGS_DEFAULT,
                       "subdir",
                       _subdir.c_str());
  DBG_INFO(0, "TransactionStorageFile successfully stored");
  return Error();
}



Error TransactionStorageFile::clear(){
  int year, month, day;
  Error err, err2;

  DBG_INFO(0, "Clearing transactions \"%s/%s\"",
           _dir.c_str(), _subdir.c_str());
  err=getFirstDay(year, month, day);
  while(err.isOk()) {
    unsigned int hdl;

    try {
      hdl=openDay(year, month, day, true);
    }
    catch(Error xerr) {
      return Error("TransactionStorageFile::clear", xerr);
    }
    err=clearDay(hdl);
    if (!err.isOk()) {
      abandonDay(hdl);
      return Error("TransactionStorageFile::clear", err);
    }
    err=closeDay(hdl);
    if (!err.isOk()) {
      return Error("TransactionStorageFile::clear", err);
    }
    err=getNextDay(year, month, day);
  } // while

  _availDays.clear();
  return Error();
}










