/***************************************************************************
  $RCSfile: accupdate.cpp,v $
  -------------------
  cvs         : $Id: accupdate.cpp,v 1.18 2004/02/29 21:30:36 rbla Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

/* Internationalization */
#ifdef ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define I18N(m) gettext(m)
#else
# define I18N(m) m
#endif


#include "accupdate.h"
#include "accountjobs.h"

#include <gwenhywfar/debug.h>


ModuleAccountUpdate::ModuleAccountUpdate()
:AqmoneyModule() {
}



ModuleAccountUpdate::~ModuleAccountUpdate(){
}



void ModuleAccountUpdate::dumpDescription(FILE *f){
  fprintf(f, "%s",
          I18N("Updates account informations (such as balance,\n"
               "   transaction list, standing orders and dated\n"
               "   transfers).\n"));
}


void ModuleAccountUpdate::_usage(){
  fprintf(stderr, "\n");
  fprintf(stderr, "%s aqmoney2 accupdate [-a %s] [-as %s] [-b %s]\n"
                  "%s   [-c %s] [-f %s] [-t %s]\n",
                  I18N("Usage:"),
                  I18N("ACCOUNTID"), I18N("ACCOUNTSUBID"), I18N("BANKCODE"),
                  I18N("      "),
                  I18N("COUNTRYCODE"), I18N("YYYYMMDD"), I18N("YYYYMMDD"));
  fprintf(stderr, "\n");
  dumpDescription(stderr);
  fprintf(stderr, "\n%s\n\n", I18N("See aqmoney2(1) for detailed help."));
}



int ModuleAccountUpdate::_checkArgs(s_args &args, int argc, char **argv){
  Error err;
  int i;
  string tmp;

  i=0;
  args.country=0;
  args.bankCode="*";
  args.accountId="*";
  args.accountSubId="*";

  while (i<argc){
    tmp=argv[i];
    if (tmp.empty()) {
      fprintf(stderr, I18N("Error: Empty command specific argument\n"));
      return -1;
    }
    if (tmp=="-c") {
      i++;
      if (i>=argc)
	return -1;
      args.country=atoi(argv[i]);
    }
    else if (tmp=="-b") {
      i++;
      if (i>=argc)
        return -1;
      args.bankCode=argv[i];
    }
    else if (tmp=="-a") {
      i++;
      if (i>=argc)
        return -1;
      args.accountId=argv[i];
    }
    else if (tmp=="-as") {
      i++;
      if (i>=argc)
        return -1;
      args.accountSubId=argv[i];
    }
    else if (tmp=="-f" || tmp=="--fromdate") {
      i++;
      if (i>=argc)
        return -1;
      args.fromDate=Date(argv[i]);
    }
    else if (tmp=="-t" || tmp=="--todate") {
      i++;
      if (i>=argc)
        return -1;
      args.toDate=Date(argv[i]);
    }
    else if (tmp=="-h" || tmp=="--help") {
      _usage();
      return -2;
    }
    else {
      if (tmp[0]=='-') {
        fprintf(stderr, I18N("Option \"%s\" is unknown to this command\n"),
                tmp.c_str());
        return -1;
      }
      args.params.push_back(argv[i]);
    }
    i++;
  } // while
  // that's it
  return 0;
}




ModuleError ModuleAccountUpdate::run(AqMoneyAPI *api,
                                     int argc, char **argv,
                                     const string &openhbciFile,
                                     const string &aqmoneyFile){
  int rv;
  Error err;
  list<Pointer<Account> > accs;
  list<Pointer<Account> >::iterator ait;
  Pointer<Outbox> queue;
  list<Pointer<AccountJob> > jobs;
  list<Pointer<AccountJob> >::iterator jit;
  Pointer<AccountJob> job;
  bool errorFlag;

  rv=_checkArgs(_args,argc,argv);
  if (rv==-2)
    return ModuleErrorSuccess;
  else if (rv==-1) {
    _usage();
    return ModuleErrorArguments;
  }

  err=api->loadEnvironment(openhbciFile, aqmoneyFile);
  if (!err.isOk()) {
    DBG_ERROR(0, "Error loading environment (%s)",
              err.errorString().c_str());
    return ModuleErrorLoad;
  }
  accs=api->getAccounts(_args.country,
                        _args.bankCode,
                        _args.accountId,
                        _args.accountSubId);
  if (accs.empty()) {
    fprintf(stderr,
            "%s",
            I18N("No matching accounts found.\n"));
    return ModuleErrorInvalidArgs;
  }

  try {
    DBG_INFO(0, "Creating new outbox");
    queue=new Outbox();

    for (ait=accs.begin(); ait!=accs.end(); ait++) {
      Pointer<Customer> cust;
      Date fromDate;

      if (!_args.fromDate.isValid()) {
        int year, month, day;

        err=(*ait).ref().storage().ref().getLastDay(year, month, day);
        if (err.isOk()) {
          char dbuffer[16];

          snprintf(dbuffer, sizeof(dbuffer), "%04d%02d%02d",
                   year, month, day);
          fromDate=Date(dbuffer);
        }
      }
      else
        fromDate=_args.fromDate;

      cust=api->findCustomer((*ait).ref().country(),
                             (*ait).ref().serverBankCode(),
                             (*ait).ref().customer());
      if (!cust.isValid()) {
        DBG_ERROR(0, "Customer \"%s\" not found, aborting",
                  (*ait).ref().customer().c_str());
        return ModuleErrorInconsistentSystem;
      }

      /* JobGetBalance */
      job=new JobGetBalance(cust, *ait);
      DBG_INFO(0, "Adding getbalance job for account %d%s%s to queue",
               (*ait).ref().country(),
               (*ait).ref().bankCode().c_str(),
               (*ait).ref().accountId().c_str());
      jobs.push_back(job);
      queue.ref().addJob(job.cast<OutboxJob>());

      /* JobGetTransactions */
      job=new JobGetTransactions(cust, *ait, fromDate, _args.toDate);
      DBG_INFO(0, "Adding GetTransactions job for account %d%s%s to queue",
               (*ait).ref().country(),
               (*ait).ref().bankCode().c_str(),
               (*ait).ref().accountId().c_str());
      jobs.push_back(job);
      queue.ref().addJob(job.cast<OutboxJob>());

      /* JobGetStandingOrders */
      job=new JobGetStandingOrders(cust, *ait);
      DBG_INFO(0, "Adding GetStandingOrders job for account %d%s%s to queue",
               (*ait).ref().country(),
               (*ait).ref().bankCode().c_str(),
               (*ait).ref().accountId().c_str());
      jobs.push_back(job);
      queue.ref().addJob(job.cast<OutboxJob>());

      /* JobGetDatedTransfers */
      job=new JobGetDatedTransfers(cust, *ait, _args.fromDate, _args.toDate);
      DBG_INFO(0, "Adding GetDatedTransfers job for account %d%s%s to queue",
               (*ait).ref().country(),
               (*ait).ref().bankCode().c_str(),
               (*ait).ref().accountId().c_str());
      jobs.push_back(job);
      queue.ref().addJob(job.cast<OutboxJob>());

    } // for

    DBG_INFO(0, "Executing queue");
    err=api->executeQueue(queue);
    if (!err.isOk()) {
      fprintf(stderr, "Error: %s\n", err.errorString().c_str());
    }
    DBG_INFO(0, "Queue executed.");

    DBG_INFO(0, "Committing results");
    err=api->processOutboxResponse(queue.ref().response());
    if (!err.isOk()) {
      DBG_ERROR(0, "Error: %s", err.errorString().c_str());
      return ModuleErrorExecution;
    }
    DBG_INFO(0, "Ok.");
    errorFlag=false;
    fprintf(stdout, I18N("Detailed job results:\n"));
    for (jit=jobs.begin(); jit!=jobs.end(); jit++) {
      bool isBad;

      isBad=(*jit).ref().result()==HBCI_JOB_RESULT_FAILED;
      fprintf(stdout, "%s",
              (*jit).ref().resultString(!isBad).c_str());

      if (!isBad) {
        err=(*jit).ref().commit();
        if (!err.isOk()) {
          DBG_ERROR(0, "Error: %s", err.errorString().c_str());
          errorFlag=true;
        }
      }
      else {
        DBG_ERROR(0, "Error in job");
        errorFlag=true;
      }
    } // for
    DBG_INFO(0, "Saving environment");
    err=api->saveEnvironment(openhbciFile, aqmoneyFile);
    if (!err.isOk()) {
      DBG_ERROR(0, "Error saving environment (%s)",
                err.errorString().c_str());
      return ModuleErrorSave;
    }
    DBG_INFO(0, "Saving environment done.");
  } // try
  catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    return ModuleErrorExecution;
  }

  if (errorFlag)
    return ModuleErrorExecution;
  return ModuleErrorSuccess;
}









