/***************************************************************************
  $RCSfile: convert.cpp,v $
  -------------------
  cvs         : $Id: convert.cpp,v 1.17 2004/02/29 21:30:36 rbla Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

/* Internationalization */
#ifdef ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define I18N(m) gettext(m)
#else
# define I18N(m) m
#endif


#include "convert.h"

#include <gwenhywfar/debug.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>


ModuleConvert::ModuleConvert()
:AqmoneyModule() {
}



ModuleConvert::~ModuleConvert(){
}



void ModuleConvert::dumpDescription(FILE *f){
  fprintf(f, "%s", I18N("Converts between different file formats\n"));
}


void ModuleConvert::_usage(){
  fprintf(stderr, "\n");
  fprintf(stderr, "%s aqmoney2 convert\n"
                  "%s   [-it db|old|dtaus|swift] [-ip %s]\n"
                  "%s   [-ot db|dtaus|csv] [-op %s] %s %s\n",
                  I18N("Usage:"),
                  I18N("      "), I18N("PARAM"),
                  I18N("      "), I18N("PARAM"), I18N("INFILE ..."), I18N("OUTFILE"));
  fprintf(stderr, "\n");
  dumpDescription(stderr);
  fprintf(stderr, "\n%s\n\n", I18N("See aqmoney2(1) for detailed help."));
}



int ModuleConvert::_checkArgs(s_args &args, int argc, char **argv){
  int i;
  string tmp;

  args.intype="db";
  args.outtype="db";

  i=0;
  while (i<argc){
    tmp=argv[i];
    if (tmp.empty()) {
      fprintf(stderr, I18N("Error: Empty command specific argument\n"));
      return -1;
    }
    if (tmp=="-it") {
      i++;
      if (i>=argc)
        return -1;
      args.intype=argv[i];
    }
    else if (tmp=="-ip") {
      i++;
      if (i>=argc)
        return -1;
      args.inparam=argv[i];
    }
    else if (tmp=="-ot") {
      i++;
      if (i>=argc)
        return -1;
      args.outtype=argv[i];
    }
    else if (tmp=="-op") {
      i++;
      if (i>=argc)
        return -1;
      args.outparam=argv[i];
    }
    else if (tmp=="-h" || tmp=="--help") {
      _usage();
      return -2;
    }
    else {
      if (tmp[0]=='-') {
        fprintf(stderr, I18N("Option \"%s\" is unknown to this command\n"),
                tmp.c_str());
        return -1;
      }
      args.params.push_back(argv[i]);
    }
    i++;
  } // while
  // that's it

  // check required arguments
  if (args.params.empty()) {
    fprintf(stderr, I18N("No files given.\n"));
    return -1;
  }
  if (args.params.size()<2) {
    fprintf(stderr, I18N("No output file given.\n"));
    return -1;
  }

  return 0;
}



ModuleError ModuleConvert::run(AqMoneyAPI *api,
                               int argc, char **argv,
                               const string &openhbciFile,
                               const string &aqmoneyFile){
  int rv;
  Error err;
  list<string>::iterator sit;
  string outfile;
  GWEN_BUFFEREDIO *bout;
  int outfd;
  Pointer<AQMConverter> outconv;
  GWEN_ERRORCODE gerr;
  GWEN_DB_NODE *transactions;

  rv=_checkArgs(_args,argc,argv);
  if (rv==-2)
    return ModuleErrorSuccess;
  else if (rv==-1) {
    _usage();
    return ModuleErrorArguments;
  }

  err=api->loadEnvironment(openhbciFile, aqmoneyFile);
  if (!err.isOk()) {
    DBG_ERROR(0, "Error loading environment (%s)",
              err.errorString().c_str());
    return ModuleErrorLoad;
  }

  outfile=_args.params.back();
  _args.params.pop_back();

  transactions=GWEN_DB_Group_new("transactions");
  // now read all files
  for (sit=_args.params.begin(); sit!=_args.params.end(); sit++) {
    int infd;
    GWEN_BUFFEREDIO *bin;
    Pointer<AQMConverter> inconv;

    DBG_NOTICE(0, "Reading from file \"%s\"", (*sit).c_str());
    infd=open((*sit).c_str(), O_RDONLY);
    if (infd==-1) {
      GWEN_DB_Group_free(transactions);
      fprintf(stderr, I18N("open(%s): %s\n"), (*sit).c_str(), strerror(errno));
      return ModuleErrorReadingfile;
    }
    bin=GWEN_BufferedIO_File_new(infd);
    GWEN_BufferedIO_SetReadBuffer(bin, 0, 1024);

    // create converters
    inconv=api->converterFactory(_args.intype, _args.inparam);
    if (!inconv.isValid()) {
      fprintf(stderr, I18N("Could not create converter \"%s\"\n"),
              _args.intype.c_str());
      GWEN_DB_Group_free(transactions);
      GWEN_BufferedIO_Abandon(bin);
      GWEN_BufferedIO_free(bin);
      return ModuleErrorInvalidArgs;
    }

    err=inconv.ref().readFromStream(bin);
    if (!err.isOk()) {
      fprintf(stderr, I18N("Error reading from file \"%s\": %s\n"),
              (*sit).c_str(), err.errorString().c_str());
      GWEN_DB_Group_free(transactions);
      GWEN_BufferedIO_Abandon(bin);
      GWEN_BufferedIO_free(bin);
      return ModuleErrorReadingfile;
    }

    GWEN_DB_AddGroupChildren(transactions, inconv.ref().transactions());

    GWEN_BufferedIO_Close(bin);
    GWEN_BufferedIO_free(bin);
  } // for

  // create output file
  outfd=open(outfile.c_str(), O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
  if (outfd==-1) {
    fprintf(stderr, I18N("open(%s): %s\n"), outfile.c_str(), strerror(errno));
    GWEN_DB_Group_free(transactions);
    return ModuleErrorWritingfile;
  }
  bout=GWEN_BufferedIO_File_new(outfd);
  GWEN_BufferedIO_SetWriteBuffer(bout, 0, 1024);

  // create output converter
  outconv=api->converterFactory(_args.outtype, _args.outparam);
  if (!outconv.isValid()) {
    fprintf(stderr, I18N("Could not create converter \"%s\"\n"),
            _args.outtype.c_str());
    GWEN_DB_Group_free(transactions);
    GWEN_BufferedIO_Abandon(bout);
    GWEN_BufferedIO_free(bout);
    return ModuleErrorInvalidArgs;
  }

  outconv.ref().setTransactions(transactions, true);

  err=outconv.ref().writeToStream(bout);
  if (!err.isOk()) {
    fprintf(stderr, I18N("Error writing to file \"%s\": %s\n"),
            outfile.c_str(), err.errorString().c_str());
    GWEN_BufferedIO_Abandon(bout);
    GWEN_BufferedIO_free(bout);
    return ModuleErrorWritingfile;
  }
  gerr=GWEN_BufferedIO_Close(bout);
  if (!GWEN_Error_IsOk(gerr)) {
    fprintf(stderr, I18N("Error closing outfile.\n"));
    GWEN_BufferedIO_free(bout);
    return ModuleErrorWritingfile;
  }
  GWEN_BufferedIO_free(bout);

  fprintf(stdout, I18N("Conversion done.\n"));
  return ModuleErrorSuccess;
}









