/***************************************************************************
  $RCSfile: createkeys.cpp,v $
  -------------------
  cvs         : $Id: createkeys.cpp,v 1.13 2004/02/29 21:30:36 rbla Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

/* Internationalization */
#ifdef ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define I18N(m) gettext(m)
#else
# define I18N(m) m
#endif



#include "adminjobs.h"
#include "createkeys.h"
#include "accountjobs.h"

#include <gwenhywfar/debug.h>


ModuleCreateKeys::ModuleCreateKeys()
:AqmoneyModule() {
}



ModuleCreateKeys::~ModuleCreateKeys(){
}



void ModuleCreateKeys::dumpDescription(FILE *f){
  fprintf(f, "%s",
          I18N("Creates private and public keys [RDH mode].\n"));
}


void ModuleCreateKeys::_usage(){
  fprintf(stderr, "\n");
  fprintf(stderr, "%s aqmoney2 createkeys [-a] [-b %s] [-c %s] [-C %s]\n",
                  I18N("Usage:"),
                  I18N("BANKCODE"), I18N("COUNTRYCODE"), I18N("CUSTOMERID"));
  fprintf(stderr, "\n");
  dumpDescription(stderr);
  fprintf(stderr, "\n%s\n\n", I18N("See aqmoney2(1) for detailed help."));
}



int ModuleCreateKeys::_checkArgs(s_args &args, int argc, char **argv){
  int i;
  string tmp;

  i=0;
  args.country=0;
  args.bankCode="*";
  args.customerId="*";
  args.activate=false;

  while (i<argc){
    tmp=argv[i];
    if (tmp.empty()) {
      fprintf(stderr, I18N("Error: Empty command specific argument\n"));
      return -1;
    }
    if (tmp=="-c") {
      i++;
      if (i>=argc)
	return -1;
      args.country=atoi(argv[i]);
    }
    else if (tmp=="-b") {
      i++;
      if (i>=argc)
        return -1;
      args.bankCode=argv[i];
    }
    else if (tmp=="-C") {
      i++;
      if (i>=argc)
        return -1;
      args.customerId=argv[i];
    }
    else if (tmp=="-a") {
      args.activate=true;
    }
    else if (tmp=="-h" || tmp=="--help") {
      _usage();
      return -2;
    }
    else {
      if (tmp[0]=='-') {
        fprintf(stderr, I18N("Option \"%s\" is unknown to this command\n"),
                tmp.c_str());
        return -1;
      }
      args.params.push_back(argv[i]);
    }
    i++;
  } // while
  // that's it
  return 0;
}




ModuleError ModuleCreateKeys::run(AqMoneyAPI *api,
                                  int argc, char **argv,
                                  const string &openhbciFile,
                                  const string &aqmoneyFile){
  int rv;
  Error err;
  list<Pointer<Customer> > custs;
  Pointer<Customer> cust;
  Pointer<Bank> bank;
  Pointer<User> user;
  Pointer<Medium> medium;
  Pointer<MediumRDHBase> mrdh;

  rv=_checkArgs(_args,argc,argv);
  if (rv==-2)
    return ModuleErrorSuccess;
  else if (rv==-1) {
    _usage();
    return ModuleErrorArguments;
  }

  err=api->loadEnvironment(openhbciFile, aqmoneyFile);
  if (!err.isOk()) {
    fprintf(stderr, I18N("Error loading environment (%s)"),
            err.errorString().c_str());
    return ModuleErrorLoad;
  }
  custs=api->getCustomers(_args.country,
                          _args.bankCode,
                          _args.customerId);
  if (custs.empty()) {
    fprintf(stderr,
            "%s",
            I18N("No matching customers found.\n"));
    return ModuleErrorInvalidArgs;
  }
  if (custs.size()>1) {
    fprintf(stderr,
            "%s",
            I18N("Multiple matching customers found, "
                 "please be more precise\n"));
    return ModuleErrorInvalidArgs;
  }

  try {
    cust=custs.front();
    user=cust.ref().user();
    bank=user.ref().bank();
    medium=user.ref().medium();

    if (medium.ref().securityMode()!=HBCI_SECURITY_RDH) {
      fprintf(stderr,
              "%s",
              I18N("Not a RDH medium\n"));
      return ModuleErrorInvalidArgs;
    }
    mrdh=medium.cast<MediumRDHBase>();

    // mount medium
    err=mrdh.ref().mountMedium();
    if (!err.isOk()) {
      fprintf(stderr,
              I18N("Error mounting medium (%s)\n"),
              err.errorString().c_str());
      return ModuleErrorExecution;
    }

    // select context
    err=mrdh.ref().selectContext(bank.ref().country(),
                                 bank.ref().bankCode(),
                                 user.ref().userId());
    if (!err.isOk()) {
      mrdh.ref().unmountMedium();
      fprintf(stderr,
              I18N("Error selecting user context (%s)\n"),
              err.errorString().c_str());
      return ModuleErrorExecution;
    }

    err=mrdh.ref().createUserKeys(_args.activate);
    if (!err.isOk()) {
      mrdh.ref().unmountMedium();
      fprintf(stderr,
              I18N("Error creating keys (%s)\n"),
              err.errorString().c_str());
      return ModuleErrorExecution;
    };

    err=mrdh.ref().unmountMedium();
    if (!err.isOk()){
      fprintf(stderr,
              I18N("Error unmounting medium (%s)\n"),
              err.errorString().c_str());
      return ModuleErrorExecution;
    }

    if (_args.activate)
      fprintf(stderr,
              I18N("Created and activated keys for user %d/%s/%s\n"),
              bank.ref().country(),
              bank.ref().bankCode().c_str(),
              user.ref().userId().c_str());
    else
      fprintf(stderr,
              I18N("Created (but not activated) keys for user %d/%s/%s\n"),
              bank.ref().country(),
              bank.ref().bankCode().c_str(),
              user.ref().userId().c_str());

    DBG_INFO(0, "Saving environment");
    err=api->saveEnvironment(openhbciFile, aqmoneyFile);
    if (!err.isOk()) {
      fprintf(stderr,
              I18N("Error saving environment (%s)"),
              err.errorString().c_str());
      return ModuleErrorSave;
    }
    DBG_INFO(0, "Saving environment done.");
  } // try

  catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    return ModuleErrorExecution;
  }

  return ModuleErrorSuccess;
}









