/***************************************************************************
  $RCSfile: getbalance.cpp,v $
  -------------------
  cvs         : $Id: getbalance.cpp,v 1.11 2004/02/29 21:30:36 rbla Exp $
  begin       : Wed Oct 15 2003
  copyright   : (C) 2003 by Martin Preuss
  email       : martin@libchipcard.de

 ***************************************************************************
 *    Please see the file COPYING in top source dir for licence details    *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

/* Internationalization */
#ifdef ENABLE_NLS
# include <libintl.h>
# include <locale.h>
# define I18N(m) gettext(m)
#else
# define I18N(m) m
#endif



#include "getbalance.h"
#include "accountjobs.h"

#include <gwenhywfar/debug.h>


ModuleGetBalance::ModuleGetBalance()
:AqmoneyModule() {
}



ModuleGetBalance::~ModuleGetBalance(){
}



void ModuleGetBalance::dumpDescription(FILE *f){
  fprintf(f, "%s",
          I18N("Retrieves the balance of your accounts\n"));
}


void ModuleGetBalance::_usage(){
  fprintf(stderr, "\n");
  fprintf(stderr, "%s aqmoney2 createkeys [-a %s] [-b %s] [-c %s]\n",
                  I18N("Usage:"),
                  I18N("ACCOUNTID"), I18N("BANKCODE"), I18N("COUNTRYCODE"));
  fprintf(stderr, "\n");
  dumpDescription(stderr);
  fprintf(stderr, "\n%s\n\n", I18N("See aqmoney2(1) for detailed help."));
}



int ModuleGetBalance::_checkArgs(s_args &args, int argc, char **argv){
  int i;
  string tmp;

  i=0;
  args.country=0;
  args.bankCode="*";
  args.accountId="*";

  while (i<argc){
    tmp=argv[i];
    if (tmp.empty()) {
      fprintf(stderr, I18N("Error: Empty command specific argument\n"));
      return -1;
    }
    if (tmp=="-c") {
      i++;
      if (i>=argc)
	return -1;
      args.country=atoi(argv[i]);
    }
    else if (tmp=="-b") {
      i++;
      if (i>=argc)
        return -1;
      args.bankCode=argv[i];
    }
    else if (tmp=="-a") {
      i++;
      if (i>=argc)
        return -1;
      args.accountId=argv[i];
    }
    else if (tmp=="-h" || tmp=="--help") {
      _usage();
      return -2;
    }
    else {
      if (tmp[0]=='-') {
        fprintf(stderr, I18N("Option \"%s\" is unknown to this command\n"),
                tmp.c_str());
        return -1;
      }
      args.params.push_back(argv[i]);
    }
    i++;
  } // while
  // that's it
  return 0;
}




ModuleError ModuleGetBalance::run(AqMoneyAPI *api,
                                  int argc, char **argv,
                                  const string &openhbciFile,
                                  const string &aqmoneyFile){
  int rv;
  Error err;
  list<Pointer<Account> > accs;
  list<Pointer<Account> >::iterator ait;
  Pointer<Outbox> queue;
  list<Pointer<JobGetBalance> > jobs;
  list<Pointer<JobGetBalance> >::iterator jit;
  Pointer<JobGetBalance> job;
  bool errorFlag;

  rv=_checkArgs(_args,argc,argv);
  if (rv==-2)
    return ModuleErrorSuccess;
  else if (rv==-1) {
    _usage();
    return ModuleErrorArguments;
  }

  err=api->loadEnvironment(openhbciFile, aqmoneyFile);
  if (!err.isOk()) {
    DBG_ERROR(0, "Error loading environment (%s)",
              err.errorString().c_str());
    return ModuleErrorLoad;
  }
  accs=api->getAccounts(_args.country,
                        _args.bankCode,
                        _args.accountId);
  if (accs.empty()) {
    fprintf(stderr,
            "%s",
            I18N("No matching accounts found.\n"));
    return ModuleErrorInvalidArgs;
  }

  try {
    DBG_INFO(0, "Creating new outbox");
    queue=new Outbox();

    for (ait=accs.begin(); ait!=accs.end(); ait++) {
      Pointer<Customer> cust;

      cust=api->findCustomer((*ait).ref().country(),
                             (*ait).ref().serverBankCode(),
                             (*ait).ref().customer());
      if (!cust.isValid()) {
        DBG_ERROR(0, "Customer \"%s\" not found, aborting",
                  (*ait).ref().customer().c_str());
        return ModuleErrorInconsistentSystem;
      }
      job=new JobGetBalance(cust, *ait);
      DBG_INFO(0, "Adding job for account %d%s%s to queue",
               (*ait).ref().country(),
               (*ait).ref().bankCode().c_str(),
               (*ait).ref().accountId().c_str());
      jobs.push_back(job);
      queue.ref().addJob(job.cast<OutboxJob>());
    } // for

    DBG_INFO(0, "Executing queue");
    err=api->executeQueue(queue);
    if (!err.isOk()) {
      fprintf(stderr, I18N("Error: %s\n"), err.errorString().c_str());
    }
    DBG_INFO(0, "Queue executed.");

    DBG_INFO(0, "Committing results");
    err=api->processOutboxResponse(queue.ref().response());
    if (!err.isOk()) {
      DBG_ERROR(0, "Error: %s", err.errorString().c_str());
      return ModuleErrorExecution;
    }
    DBG_INFO(0, "Ok.");
    errorFlag=false;
    fprintf(stdout, I18N("Detailed job results:\n"));
    for (jit=jobs.begin(); jit!=jobs.end(); jit++) {
      bool isBad;

      isBad=(*jit).ref().result()==HBCI_JOB_RESULT_FAILED;
      fprintf(stdout, "%s",
              (*jit).ref().resultString(!isBad).c_str());

      if (!isBad) {
        err=(*jit).ref().commit();
        if (!err.isOk()) {
          DBG_ERROR(0, "Error: %s", err.errorString().c_str());
          errorFlag=true;
        }
      }
      else {
        DBG_ERROR(0, "Error in job");
        errorFlag=true;
      }
    } // for
    DBG_INFO(0, "Saving environment");
    err=api->saveEnvironment(openhbciFile, aqmoneyFile);
    if (!err.isOk()) {
      DBG_ERROR(0, "Error saving environment (%s)",
                err.errorString().c_str());
      return ModuleErrorSave;
    }
    DBG_INFO(0, "Saving environment done.");
  } // try
  catch (Error xerr) {
    DBG_ERROR(0, "Error: %s", xerr.errorString().c_str());
    return ModuleErrorExecution;
  }

  if (errorFlag)
    return ModuleErrorExecution;
  return ModuleErrorSuccess;
}









